import os
from tensorflow.keras.utils import get_file
from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.layers import ReLU
from tensorflow.keras.layers import Input
from tensorflow.keras.layers import Conv2DTranspose
from tensorflow.keras.layers import UpSampling2D
from tensorflow.keras.layers import Add
from tensorflow.keras.layers import concatenate
from tensorflow.keras.layers import ZeroPadding2D
from tensorflow.keras import backend as K
from tensorflow.keras.models import Model


WEIGHT_PATH = ('https://github.com/oarriaga/altamira-data/releases/download'
               '/v0.10/HigherHRNet.hdf5')


def stem(tensor, filters):
    x = ZeroPadding2D(padding=(1, 1), name='pad')(tensor)
    x = Conv2D(filters, 3, strides=2, use_bias=False, name='conv1')(x)
    x = BatchNormalization(momentum=0.1, epsilon=1e-05,
                           name='bn1')(x, training=False)
    x = ReLU(name='relu')(x)
    x = ZeroPadding2D(padding=(1, 1), name='pad_1')(x)
    x = Conv2D(filters, 3, strides=2, use_bias=False, name='conv2')(x)
    x = BatchNormalization(momentum=0.1, epsilon=1e-05, name='bn2')(x)
    x = ReLU()(x)
    return x


def bottleneck(tensor, filters, expansion, downsample=None, name=None):
    residual = tensor
    x = Conv2D(filters, 1, use_bias=False, name=name + '.conv1')(tensor)
    x = BatchNormalization(momentum=0.1, epsilon=1e-05, name=name + '.bn1')(x)
    x = ReLU(name=name + '.relu')(x)
    x = Conv2D(filters, 3, padding='same',
               use_bias=False, name=name + '.conv2')(x)
    x = BatchNormalization(momentum=0.1, epsilon=1.0e-5, name=name + '.bn2')(x)
    x = ReLU()(x)
    x = Conv2D(filters * expansion, 1, use_bias=False, name=name + '.conv3')(x)
    x = BatchNormalization(momentum=0.1, epsilon=1.0e-5, name=name + '.bn3')(x)
    if downsample is not None:
        x1 = Conv2D(256, 1, use_bias=False,
                    name='layer1.0.downsample.0')(tensor)
        residual = BatchNormalization(momentum=0.1, epsilon=1e-05,
                                      name='layer1.0.downsample.1')(x1)
    x = Add()([x, residual])
    x = ReLU()(x)
    return x


def basic_block(tensor, filters, name=None):
    residual = tensor
    x = Conv2D(filters, 3, padding='same',
               use_bias=False, name=name + '.conv1')(tensor)
    x = BatchNormalization(momentum=0.1, epsilon=1e-05, name=name + '.bn1')(x)
    x = ReLU(name=name + '.relu')(x)
    x = Conv2D(filters, 3, padding='same',
               use_bias=False, name=name + '.conv2')(x)
    x = BatchNormalization(momentum=0.1, epsilon=1e-05, name=name + '.bn2')(x)
    x = Add()([x, residual])
    x = ReLU()(x)
    return x


def transition_block(tensor, alpha, name):
    in_channels = K.int_shape(tensor)[-1]
    if in_channels == 256:
        filters = 32 * alpha
    else:
        filters = in_channels * alpha
    x = ZeroPadding2D(padding=(1, 1))(tensor)
    x = Conv2D(filters, 3, strides=2, use_bias=False, name=name + '0.0')(x)
    x = BatchNormalization(momentum=0.1, epsilon=1e-05, name=name + '0.1')(x)
    x = ReLU()(x)
    return x


def blocks_in_branch(tensors, stage, in_channels, name):
    if stage != len(tensors):
        raise ValueError('''outputs {} feed to fuse_layers must to be same as
                         num_branches {}'''.format(tensors, stage))
    for arg in range(stage):
        filters = in_channels * (2 ** arg)
        tensors[arg] = basic_block(tensors[arg], filters,
                                   name=name[:18] + str(arg) + '.' + name[18:])
    return tensors


def final_layers(num_keypoints, with_AE_loss=None, num_deconv=1):
    final_layers = []
    if with_AE_loss[0]:
        output_channels = num_keypoints * 2
    else:
        output_channels = num_keypoints
    x = Conv2D(output_channels, 1, padding='same', name='final_layers.0')
    final_layers.append(x)

    for arg in range(num_deconv):
        if with_AE_loss[arg + 1]:
            output_channels = num_keypoints * 2
        else:
            output_channels = num_keypoints
        x1 = Conv2D(output_channels, 1, padding='same', name='final_layers.1')
        final_layers.append(x1)
    return final_layers


def deconv_layers(tensor, output_channels, num_deconv=1):
    for arg in range(num_deconv):
        x = Conv2DTranspose(output_channels, 4, strides=2,
                            padding='same', use_bias=False,
                            name='deconv_layers.0.0.0')(tensor)
        x = BatchNormalization(momentum=0.1, epsilon=1e-05,
                               name='deconv_layers.0.0.1')(x)
        x = ReLU()(x)
        for block in range(4):
            x = basic_block(
                x, output_channels,
                name='deconv_layers.0.' + str(block + 1) + '.' + '0')
    return x


def get_names(name, branch_arg, stage_arg, counter, iterations=0):
    name1 = '.'.join((name, str(branch_arg), str(stage_arg),
                     str(iterations + counter)))
    name2 = '.'.join((name, str(branch_arg), str(stage_arg),
                     str(iterations + counter + .1)))
    return [name1, name2]


def fuse_layers(tensors, stage, output_branches, filters=32, name=None):
    if stage != len(tensors):
        raise ValueError('''outputs {} feed to fuse_layers must to be same as
                         num_branches {}'''.format(tensors, stage))
    all_tensors = []
    for branch_arg in range(output_branches):
        x_to_y_tensors = []
        for stage_arg in range(stage):
            # step: how much the feature map is upsampled or downsampled
            steps = stage_arg - branch_arg
            if steps == 0:
                y = tensors[branch_arg]

            elif steps > 0:  # upsample
                name0 = '.'.join((name, str(branch_arg), str(stage_arg)))
                y = upsample(tensors[stage_arg], filters * (2 ** branch_arg),
                             size=(2**steps, 2**steps), name=name0)

            elif steps < 0:  # downsample
                y_flag = False
                iterations = 0
                for k in range((-1 * steps) - 1):
                    iterations += 1
                    if y_flag:
                        name1 = get_names(name, branch_arg, stage_arg, 1.0)
                        y = downsample(y, filters * (2 ** stage_arg), name1)

                    else:
                        name2 = get_names(name, branch_arg, stage_arg, 0.0)
                        y = downsample(tensors[stage_arg],
                                       filters * (2 ** stage_arg), name2,
                                       with_padding=False)
                    y = ReLU()(y)
                    y_flag = True

                if not y_flag:
                    tensors[stage_arg] = ZeroPadding2D()(tensors[stage_arg])
                    name3 = get_names(name, branch_arg, stage_arg, 0.0)
                    y = downsample(tensors[stage_arg],
                                   filters * (2 ** branch_arg), name3,
                                   with_padding=False)

                else:
                    name4 = get_names(name, branch_arg, stage_arg,
                                      .0, iterations)
                    y = downsample(y, filters * (2 ** branch_arg), name4)
            x_to_y_tensors.append(y)

        all_tensors.append(x_to_y_tensors)

    x_fused = []
    for x_tensor_arg in range(len(all_tensors)):
        for y_tensor_arg in range(len(x_to_y_tensors)):
            if y_tensor_arg == 0:
                x_fused.append(all_tensors[x_tensor_arg][0])
            else:
                x = Add()([x_fused[x_tensor_arg],
                          all_tensors[x_tensor_arg][y_tensor_arg]])
                x_fused[x_tensor_arg] = x

    for x_fused_arg in range(len(x_fused)):
        x_fused[x_fused_arg] = ReLU()(x_fused[x_fused_arg])
    return x_fused


def upsample(tensor, filters, size, name=None):
    x = Conv2D(filters, 1, use_bias=False, name=name + '.0')(tensor)
    x = BatchNormalization(momentum=0.1, epsilon=1e-05, name=name + '.1')(x)
    x = UpSampling2D(size=size, interpolation='nearest', name=name + '.2')(x)
    return x


def downsample(tensor, filters, name=None, with_padding=True):
    if with_padding:
        tensor = ZeroPadding2D(padding=(1, 1))(tensor)
    x = Conv2D(filters, 3, strides=2, use_bias=False, name=name[0])(tensor)
    x = BatchNormalization(momentum=0.1, epsilon=1e-05, name=name[1])(x)
    return x


def HigherHRNet(weights='COCO', input_shape=(None, None, 3), num_keypoints=17,
                with_AE_loss=[True, False]):
    """Human pose estimation detector for any input size of images.
    # Arguments
        weights: String or None. If string should be a valid dataset name.
            Current valid datasets include `COCO`.
        input_shape: List of integers. Input shape to the model including only
            spatial and channel resolution e.g. (512, 512, 3).
        num_keypoints: Int. Number of joints.
        with_AE_loss: List of boolean.

    # Reference
        - [HigherHRNet: Scale-Aware Representation Learning for Bottom-Up
           Human Pose Estimation](https://arxiv.org/abs/1908.10357)
    """

    image = Input(shape=input_shape, name='image')
    x = stem(image, 64)
    # print(f"check 1 stem TF ==> {x.get_shape()}")

    # First group of bottleneck (resnet) modules
    # Stage 1 -----------------------------------------------------------------
    x = bottleneck(x, filters=64, expansion=4,
                   downsample=True, name='layer1' + '.0')
    for block in range(3):
        x = bottleneck(x, filters=64, expansion=4,
                       downsample=None, name='layer1' + '.' + str(block + 1))

    x_list = []
    # Creation of the first two branches (one full and one half resolution)
    x1 = Conv2D(32, 3, strides=1, padding='same',
                use_bias=False, name='transition1.0.0')(x)
    x1 = BatchNormalization(momentum=0.1, name='transition1.0.1')(x1)
    x1 = ReLU()(x1)

    x_list.append(x1)
    x_list.append(transition_block(x, 2, name='transition1.1.'))

    # Stage 2 -----------------------------------------------------------------
    for block in range(4):
        x_list = blocks_in_branch(x_list, stage=2, in_channels=32,
                                  name='stage2.0.branches.' + str(block))
    x_list = fuse_layers(x_list, stage=2, output_branches=2,
                         name='stage2.0.fuse_layers')
    x_list.append(transition_block(x_list[1], 2, name='transition2.2.'))

    # Stage 3 -----------------------------------------------------------------
    for module in range(4):
        for block in range(4):
            name = 'stage3.' + str(module) + '.branches.' + str(block)
            x_list = blocks_in_branch(x_list, stage=3,
                                      in_channels=32, name=name)
        x_list = fuse_layers(x_list, stage=3, output_branches=3,
                             name='stage3.' + str(module) + '.fuse_layers')
    x_list.append(transition_block(x_list[2], 2, name='transition3.3.'))

    # Stage 4 -----------------------------------------------------------------
    for module in range(3):
        for block in range(4):
            name = 'stage4.' + str(module) + '.branches.' + str(block)
            x_list = blocks_in_branch(x_list, stage=4,
                                      in_channels=32, name=name)
        if module == 2:
            name = 'stage4.' + str(module) + '.fuse_layers'
            x_list = fuse_layers(x_list, stage=4, output_branches=1, name=name)
        else:
            name = 'stage4.' + str(module) + '.fuse_layers'
            x_list = fuse_layers(x_list, stage=4, output_branches=4, name=name)

    final_outputs = []
    x2 = x_list[0]
    output = final_layers(num_keypoints, with_AE_loss=with_AE_loss)[0](x2)
    final_outputs.append(output)

    x2 = concatenate((x2, output), -1)
    x2 = deconv_layers(x2, 32)
    x2 = final_layers(num_keypoints, with_AE_loss=with_AE_loss)[1](x2)
    final_outputs.append(x2)

    model = Model(image, outputs=final_outputs, name='HigherHRNet')

    if(weights == 'COCO'):
        URL = ('https://github.com/oarriaga/altamira-data/releases/download'
               '/v0.10/HigherHRNet_weights.hdf5')
        filename = os.path.basename(URL)
        weights_path = get_file(filename, URL, cache_subdir='paz/models')
        print('==> Loading %s model weights' % weights_path)
        model.load_weights(weights_path)
    return model



from paz import processors as pr
from paz.models import IKNet
from paz.datasets import MPIIHandJoints
from paz.backend.keypoints import flip_along_x_axis


class IKNetHandJointAngles(pr.Processor):
    """Estimate absolute and relative joint angle for the minimal hand joints
       using the 3D keypoint locations.

    # Arguments
        links_origin: Array. Reference pose of the minimal hand joints.
        parent: List. Parents of the keypoints from kinematic chain
        right_hand: Boolean. If 'True', estimate angles for right hand, else
                    estimate angles for left hand.
        keypoints3D: Array [num_joints, 3]. 3D location of keypoints.

    # Returns
        absolute_angles: Array [num_joints, 4]. quaternion repesentation
        relative_angles: Array [num_joints, 3]. axis-angle repesentation
    """
    def __init__(self, links_origin=MPIIHandJoints.links_origin,
                 parents=MPIIHandJoints.parents, right_hand=False):
        super(IKNetHandJointAngles, self).__init__()
        self.calculate_orientation = pr.ComputeOrientationVector(parents)
        self.links_origin = links_origin
        self.right_hand = right_hand
        if self.right_hand:
            self.links_origin = flip_along_x_axis(self.links_origin)
        self.links_delta = self.calculate_orientation(self.links_origin)
        self.concatenate = pr.Concatenate(0)
        self.compute_absolute_angles = pr.SequentialProcessor(
            [pr.ExpandDims(0), IKNet(), pr.Squeeze(0)])
        self.compute_relative_angles = pr.CalculateRelativeAngles()
        self.wrap = pr.WrapOutput(['absolute_angles', 'relative_angles'])

    def call(self, keypoints3D):
        delta = self.calculate_orientation(keypoints3D)
        pack = self.concatenate(
            [keypoints3D, delta, self.links_origin, self.links_delta])
        absolute_angles = self.compute_absolute_angles(pack)
        relative_angles = self.compute_relative_angles(absolute_angles)
        return self.wrap(absolute_angles, relative_angles)




import os
from tensorflow.keras.utils import get_file
from tensorflow.keras.models import Model
from tensorflow.keras.initializers import HeNormal
from tensorflow.keras.constraints import MaxNorm
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.layers import ReLU
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import Dropout
from tensorflow.keras.layers import Input
from tensorflow.keras.layers import Reshape


def dense_block(input_x, num_keypoints, rate):
    """Make a bi-linear block with optional residual connection
    # Arguments
        input_x: the batch that enters the block
        num_keypoints: integer. The size of the linear units
        rate: float [0,1]. Probability of dropping something out

    # Returns
        x: the batch after it leaves the block
    """
    kwargs = {'kernel_initializer': HeNormal(), 'bias_initializer': HeNormal(),
              'kernel_constraint': MaxNorm(max_value=1)}
    x = Dense(num_keypoints, **kwargs)(input_x)
    x = BatchNormalization()(x)
    x = ReLU()(x)
    x = Dropout(rate)(x)
    x = Dense(num_keypoints, **kwargs, )(x)
    x = BatchNormalization()(x)
    x = ReLU()(x)
    x = Dropout(rate)(x)
    x = (x + input_x)
    return x


def SimpleBaseline(input_shape=(32,), num_keypoints=16, keypoints_dim=3,
                   hidden_dim=1024, num_layers=2, rate=1, weights='human36m'):
    """Model that predicts 3D keypoints from 2D keypoints
    # Arguments
        num_keypoints: numer of kepoints
        keypoints_dim: dimension of keypoints
        hidden_dim: size of hidden layers
        input_shape: size of the input
        num_layers: number of layers
        rate: dropout drop rate

    # Returns
        keypoints3D estimation model
    """
    inputs = Input(shape=input_shape)
    kwargs = {'kernel_initializer': HeNormal(), 'bias_initializer': HeNormal(),
              'kernel_constraint': MaxNorm(max_value=1)}
    x = Dense(hidden_dim, use_bias=True, **kwargs)(inputs)
    x = BatchNormalization()(x)
    x = ReLU()(x)
    x = Dropout(rate)(x)
    for layer in range(num_layers):
        x = dense_block(x, hidden_dim, rate)
    x = Dense(num_keypoints * keypoints_dim, **kwargs)(x)
    x = Reshape((num_keypoints, keypoints_dim))(x)
    model = Model(inputs, outputs=x)
    if weights == 'human36m':
        URL = ('https://github.com/oarriaga/altamira-data/releases/download/'
               'v0.17/SIMPLE-BASELINES.hdf5')
        filename = os.path.basename(URL)
        weights_path = get_file(filename, URL, cache_subdir='paz/models')
        model.load_weights(weights_path)
    return model



from ..abstract import SequentialProcessor, Processor
from .. import processors as pr
from ..backend.image import resize_image, BILINEAR
from ..backend.keypoints import normalize_keypoints2D


class PredictRGBMask(SequentialProcessor):
    """Predicts RGB mask from a segmentation model
    # Arguments
        model: Keras segmentation model.
        epsilon: Float. Values below this value would be replaced by 0.
    """
    def __init__(self, model, epsilon=0.15):
        super(PredictRGBMask, self).__init__()
        self.add(pr.ResizeImage(model.input_shape[1:3]))
        self.add(pr.NormalizeImage())
        self.add(pr.ExpandDims(0))
        self.add(pr.Predict(model))
        self.add(pr.Squeeze(0))
        self.add(pr.ReplaceLowerThanThreshold(epsilon))
        self.add(pr.DenormalizeImage())
        self.add(pr.CastImage('uint8'))


class RGBMaskToObjectPoints3D(SequentialProcessor):
    """Predicts 3D keypoints from an RGB mask.
    # Arguments
        object_sizes: Array (3) determining the (width, height, depth)
    """
    def __init__(self, object_sizes):
        super(RGBMaskToObjectPoints3D, self).__init__()
        self.add(pr.GetNonZeroValues())
        self.add(pr.ImageToNormalizedDeviceCoordinates())
        self.add(pr.Scale(object_sizes / 2.0))


class RGBMaskToImagePoints2D(SequentialProcessor):
    """Predicts 2D image keypoints from an RGB mask.
    """
    def __init__(self):
        super(RGBMaskToImagePoints2D, self).__init__()
        self.add(pr.GetNonZeroArguments())
        self.add(pr.ArgumentsToImageKeypoints2D())


class Pix2Points(Processor):
    """Predicts RGB_mask and corresponding points2D and points3D.

    # Arguments
        model: Keras segmentation model.
        object_sizes: Array (3) determining the (width, height, depth)
        epsilon: Float. Values below this value would be replaced by 0.
        resize: Boolean. If True RGB mask is resized to original shape.
        method: Interpolation method to use if resize is True.

    # Note
        Compare with and without RGB interpolation.
    """
    def __init__(self, model, object_sizes, epsilon=0.15,
                 resize=False, method=BILINEAR):
        self.model = model
        self.resize = resize
        self.method = method
        self.object_sizes = object_sizes
        self.predict_RGBMask = PredictRGBMask(model, epsilon)
        self.mask_to_points3D = RGBMaskToObjectPoints3D(self.object_sizes)
        self.mask_to_points2D = RGBMaskToImagePoints2D()
        self.wrap = pr.WrapOutput(['points2D', 'points3D', 'RGB_mask'])

    def call(self, image):
        RGB_mask = self.predict_RGBMask(image)
        if self.resize:
            H, W, num_channels = image.shape
            RGB_mask = resize_image(RGB_mask, (W, H), self.method)
        else:
            H, W = self.model.output_shape[1:3]
        points3D = self.mask_to_points3D(RGB_mask)
        points2D = self.mask_to_points2D(RGB_mask)
        points2D = normalize_keypoints2D(points2D, H, W)
        return self.wrap(points2D, points3D, RGB_mask)



import numpy as np


class UnsolvableMatrix(Exception):
    """
    Exception raised for unsolvable matrices
    """
    pass


class DISALLOWED_OBJ(object):
    pass


DISALLOWED = DISALLOWED_OBJ()
DISALLOWED_PRINTVAL = "D"


def get_cover_matrix(shape):
    """Returns the initialized row and column cover matrix.

    # Arguments
        shape: Tuple. Shape of the cover matrix.
    """
    row_covered = np.zeros(shape, dtype='bool')
    col_covered = np.zeros(shape, dtype='bool')
    return row_covered, col_covered


def find_uncovered_zero(n, cost_matrix, row_covered, col_covered, i0, j0):
    row = -1
    col = -1
    done = False
    for row_arg in range(i0, n):
        for col_arg in range(j0, n):
            if (cost_matrix[row_arg][col_arg] == 0) and \
                    (not row_covered[row_arg]) and \
                    (not col_covered[col_arg]):
                row = row_arg
                col = col_arg
                done = True
        if done:
            break
    return (row, col)


def find_star_in_row(n, row_arg, marked):
    col = -1
    for col_arg in range(n):
        if marked[row_arg][col_arg] == 1:
            col = col_arg
            break
    return col


def find_star_in_col(n, col_arg, marked):
    row = -1
    for row_arg in range(n):
        if marked[row_arg][col_arg] == 1:
            row = row_arg
            break
    return row


def find_prime_in_row(n, row_arg, marked):
    col = -1
    for col_arg in range(n):
        if marked[row_arg][col_arg] == 2:
            col = col_arg
            break
    return col


def get_min_value(series):
    values = []
    for x in series:
        if type(x) is not type(DISALLOWED):
            values.append(x)
    if len(values) == 0:
        raise UnsolvableMatrix("One row is entirely DISALLOWED.")
    min_value = np.min(values)
    return min_value


def find_smallest_uncovered(n, row_covered, col_covered, cost_matrix):
    minval = np.inf
    for i in range(n):
        for j in range(n):
            if (not row_covered[i]) and (not col_covered[j]):
                if cost_matrix[i][j] is not DISALLOWED and \
                        minval > cost_matrix[i][j]:
                    minval = cost_matrix[i][j]
    return minval



import numpy as np
from tensorflow.keras.utils import get_file

from .. import processors as pr
from ..abstract import Processor, SequentialProcessor, Pose6D
from ..models import UNET_VGG16
from ..backend.image.draw import draw_points2D, points3D_to_RGB
from ..backend.standard import append_lists
from ..backend.keypoints import (
    translate_points2D_origin, denormalize_keypoints2D)

from .masks import Pix2Points
from .detection import HaarCascadeFrontalFace
from .keypoints import FaceKeypointNet2D32
from .detection import SSD300FAT, PostprocessBoxes2D


class EstimatePoseKeypoints(Processor):
    def __init__(self, detect, estimate_keypoints, camera, offsets,
                 model_points, class_to_dimensions, radius=3, thickness=1):
        """Pose estimation pipeline using keypoints.

        # Arguments
            detect: Function that outputs a dictionary with a key
                ``boxes2D`` having a list of ``Box2D`` messages.
            estimate_keypoints: Function that outputs a dictionary
                with a key ``keypoints`` with numpy array as value
            camera: Instance of ``paz.backend.camera.Camera`` with
                camera intrinsics.
            offsets: List of floats indicating the scaled offset to
                be added to the ``Box2D`` coordinates.
            model_points: Numpy array of shape ``(num_keypoints, 3)``
                indicating the 3D coordinates of the predicted keypoints
                from the ``esimate_keypoints`` function.
            class_to_dimensions: Dictionary with keys being the class labels
                of the predicted ``Box2D`` messages and the values a list of
                three integers indicating the width, height and depth of the
                object e.g. {'PowerDrill': [30, 20, 10]}.
            radius: Int. radius of keypoint to be drawn.
            thickness: Int. thickness of 3D box.

        # Returns
            A function that takes an RGB image and outputs the following
            inferences as keys of a dictionary:
                ``image``, ``boxes2D``, ``keypoints`` and ``poses6D``.

        """
        super(EstimatePoseKeypoints, self).__init__()
        self.num_keypoints = estimate_keypoints.num_keypoints
        self.detect = detect
        self.estimate_keypoints = estimate_keypoints
        self.square = SequentialProcessor()
        self.square.add(pr.SquareBoxes2D())
        self.square.add(pr.OffsetBoxes2D(offsets))
        self.clip = pr.ClipBoxes2D()
        self.crop = pr.CropBoxes2D()
        self.change_coordinates = pr.ChangeKeypointsCoordinateSystem()
        self.solve_PNP = pr.SolvePNP(model_points, camera)
        self.draw_keypoints = pr.DrawKeypoints2D(self.num_keypoints, radius)
        self.draw_box = pr.DrawBoxes3D(camera, class_to_dimensions,
                                       thickness=thickness)
        self.wrap = pr.WrapOutput(['image', 'boxes2D', 'keypoints', 'poses6D'])

    def call(self, image):
        boxes2D = self.detect(image)['boxes2D']
        boxes2D = self.square(boxes2D)
        boxes2D = self.clip(image, boxes2D)
        cropped_images = self.crop(image, boxes2D)
        poses6D, keypoints2D = [], []
        for cropped_image, box2D in zip(cropped_images, boxes2D):
            keypoints = self.estimate_keypoints(cropped_image)['keypoints']
            keypoints = self.change_coordinates(keypoints, box2D)
            pose6D = self.solve_PNP(keypoints)
            image = self.draw_keypoints(image, keypoints)
            image = self.draw_box(image, pose6D)
            keypoints2D.append(keypoints)
            poses6D.append(pose6D)
        return self.wrap(image, boxes2D, keypoints2D, poses6D)


class HeadPoseKeypointNet2D32(EstimatePoseKeypoints):
    """Head pose estimation pipeline using a ``HaarCascade`` face detector
        and a pre-trained ``KeypointNet2D`` estimation model.

        # Arguments
            camera: Instance of ``paz.backend.camera.Camera`` with
                camera intrinsics.
            offsets: List of floats indicating the scaled offset to
                be added to the ``Box2D`` coordinates.
            radius: Int. radius of keypoint to be drawn.

        # Example
            ``` python
            from paz.pipelines import HeadPoseKeypointNet2D32

            estimate_pose = HeadPoseKeypointNet2D32()

            # apply directly to an image (numpy-array)
            inferences = estimate_pose(image)
            ```

        # Returns
            A function that takes an RGB image and outputs the following
            inferences as keys of a dictionary:
                ``image``, ``boxes2D``, ``keypoints`` and ``poses6D``.
        """
    def __init__(self, camera, offsets=[0, 0], radius=5, thickness=2):
        detect = HaarCascadeFrontalFace(draw=False)
        estimate_keypoints = FaceKeypointNet2D32(draw=False)
        """
                               4--------1
                              /|       /|
                             / |      / |
                            3--------2  |
                            |  8_____|__5
                            | /      | /
                            |/       |/
                            7--------6

                   Z (depth)
                  /
                 /_____X (width)
                 |
                 |
                 Y (height)
        """
        KEYPOINTS3D = np.array([
            [-220, 1138, 678],  # left--center-eye
            [+220, 1138, 678],  # right-center-eye
            [-131, 1107, 676],  # left--eye close to nose
            [-294, 1123, 610],  # left--eye close to ear
            [+131, 1107, 676],  # right-eye close to nose
            [+294, 1123, 610],  # right-eye close to ear
            [-106, 1224, 758],  # left--eyebrow close to nose
            [-375, 1208, 585],  # left--eyebrow close to ear
            [+106, 1224, 758],  # right-eyebrow close to nose
            [+375, 1208, 585],  # right-eyebrow close to ear
            [0.0, 919, 909],  # nose
            [-183, 683, 691],  # lefty-lip
            [+183, 683, 691],  # right-lip
            [0.0, 754, 826],  # up---lip
            [0.0, 645, 815],  # down-lip
        ])
        KEYPOINTS3D = KEYPOINTS3D - np.mean(KEYPOINTS3D, axis=0)
        super(HeadPoseKeypointNet2D32, self).__init__(
            detect, estimate_keypoints, camera, offsets,
            KEYPOINTS3D, {None: [900, 1200, 800]}, radius, thickness)


class SingleInstancePIX2POSE6D(Processor):
    """Predicts a single pose6D from an image. Optionally if a box2D message is
        given it translates the predicted points2D to new origin located at
        box2D top-left corner.

    # Arguments
        model: Keras segmentation model.
        object_sizes: Array (3) determining the (width, height, depth)
        camera: PAZ Camera with intrinsic matrix.
        epsilon: Float. Values below this value would be replaced by 0.
        resize: Boolean. If True RGB mask is resized before computing PnP.
        class_name: Str indicating object name.
        draw: Boolean. If True drawing functions are applied to output image.

    # Returns
        Dictionary with inferred points2D, points3D, pose6D and image.
    """
    def __init__(self, model, object_sizes, camera,
                 epsilon=0.15, resize=False, class_name=None, draw=True):
        super(SingleInstancePIX2POSE6D, self).__init__()
        self.camera = camera
        self.pix2points = Pix2Points(model, object_sizes, epsilon, resize)
        self.solvePnP = pr.SolveChangingObjectPnPRANSAC(self.camera.intrinsics)
        self.draw_pose6D = pr.DrawPose6D(object_sizes, self.camera.intrinsics)
        self.wrap = pr.WrapOutput(['image', 'points2D', 'points3D', 'pose6D'])
        self.class_name = str(class_name)
        self.object_sizes = object_sizes
        self.draw = draw

    def call(self, image, box2D=None):
        inferences = self.pix2points(image)
        points2D = inferences['points2D']
        points3D = inferences['points3D']
        points2D = denormalize_keypoints2D(points2D, *image.shape[:2])
        if box2D is not None:
            points2D = translate_points2D_origin(points2D, box2D.coordinates)
            self.class_name = box2D.class_name
        pose6D = None
        if len(points3D) > self.solvePnP.MIN_REQUIRED_POINTS:
            success, R, T = self.solvePnP(points3D, points2D)
            if success:
                pose6D = Pose6D.from_rotation_vector(R, T, self.class_name)
        if (self.draw and (box2D is None) and (pose6D is not None)):
            colors = points3D_to_RGB(points3D, self.object_sizes)
            image = draw_points2D(image, points2D, colors)
            image = self.draw_pose6D(image, pose6D)
        inferences = self.wrap(image, points2D, points3D, pose6D)
        return inferences


class MultiInstancePIX2POSE6D(Processor):
    """Predicts poses6D of multiple instances the same object from an image.

    # Arguments
        estimate_pose: Function that takes as input an image and outputs a
            dictionary with points2D, points3D and pose6D messages e.g
            SingleInstancePIX2POSE6D
        offsets: List of length two containing floats e.g. (x_scale, y_scale)
        camera: PAZ Camera with intrinsic matrix.
        draw: Boolean. If True drawing functions are applied to output image.

    # Returns
        Dictionary with inferred boxes2D, poses6D and image.
    """
    def __init__(self, estimate_pose, offsets, camera=None, draw=True):
        super(MultiInstancePIX2POSE6D, self).__init__()
        self.draw = draw
        self.estimate_pose = estimate_pose
        self.object_sizes = self.estimate_pose.object_sizes
        self.camera = self.estimate_pose.camera if camera is None else camera
        valid_names = [self.estimate_pose.class_name]
        self.postprocess_boxes = PostprocessBoxes2D(offsets, valid_names)

        self.append_values = pr.AppendValues(
            ['pose6D', 'points2D', 'points3D'])
        self.clip = pr.ClipBoxes2D()
        self.crop = pr.CropBoxes2D()
        self.draw_RGBmask = pr.DrawRGBMasks(self.object_sizes)
        self.draw_boxes2D = pr.DrawBoxes2D(valid_names, colors=[[0, 255, 0]])
        self.draw_poses6D = pr.DrawPoses6D(
            self.object_sizes, camera.intrinsics)
        self.wrap = pr.WrapOutput(['image', 'boxes2D', 'poses6D'])

    def call(self, image, boxes2D):
        boxes2D = self.postprocess_boxes(boxes2D)
        boxes2D = self.clip(image, boxes2D)
        cropped_images = self.crop(image, boxes2D)
        poses6D, points2D, points3D = [], [], []
        for crop, box2D in zip(cropped_images, boxes2D):
            inferences = self.estimate_pose(crop, box2D)
            self.append_values(inferences, [poses6D, points2D, points3D])
        if self.draw:
            image = self.draw_boxes2D(image, boxes2D)
            image = self.draw_RGBmask(image, points2D, points3D)
            image = self.draw_poses6D(image, poses6D)
        return self.wrap(image, boxes2D, poses6D)


class SinglePowerDrillPIX2POSE6D(SingleInstancePIX2POSE6D):
    """Predicts the pose6D of the YCB 035_power_drill object from an image.
        Optionally if a box2D message is given it translates the predicted
        points2D to new origin located at box2D top-left corner.

    # Arguments
        camera: PAZ Camera with intrinsic matrix.
        epsilon: Float. Values below this value would be replaced by 0.
        resize: Boolean. If True RGB mask is resized before computing PnP.
        draw: Boolean. If True drawing functions are applied to output image.

    # Returns
        Dictionary with inferred points2D, points3D, pose6D and image.
    """
    def __init__(self, camera, epsilon=0.15, resize=False, draw=True):
        model = UNET_VGG16(3, (128, 128, 3))
        URL = ('https://github.com/oarriaga/altamira-data/'
               'releases/download/v0.13/')
        name = 'UNET-VGG16_POWERDRILL_weights.hdf5'
        weights_path = get_file(name, URL + name, cache_subdir='paz/models')
        print('Loading %s model weights' % weights_path)
        model.load_weights(weights_path)
        object_sizes = np.array([1840, 1870, 520]) / 10000
        class_name = '035_power_drill'
        super(SinglePowerDrillPIX2POSE6D, self).__init__(
            model, object_sizes, camera, epsilon, resize, class_name, draw)


class MultiPowerDrillPIX2POSE6D(MultiInstancePIX2POSE6D):
    """Predicts poses6D of multiple instances the YCB 035_power_drill object
        from an image.

    # Arguments
        camera: PAZ Camera with intrinsic matrix.
        offsets: List of length two containing floats e.g. (x_scale, y_scale)
        epsilon: Float. Values below this value would be replaced by 0.
        resize: Boolean. If True RGB mask is resized before computing PnP.
        draw: Boolean. If True drawing functions are applied to output image.

    # Returns
        Dictionary with inferred boxes2D, poses6D and image.
    """
    def __init__(self, camera, offsets, epsilon=0.15, resize=False, draw=True):
        estimate_pose = SinglePowerDrillPIX2POSE6D(
            camera, epsilon, resize, draw=False)
        super(MultiPowerDrillPIX2POSE6D, self).__init__(
            estimate_pose, offsets, camera, draw)


class PIX2POSEPowerDrill(Processor):
    """PIX2POSE inference pipeline with SSD300 trained on FAT and UNET-VGG16
        trained with domain randomization for the YCB object 035_power_drill.

    # Arguments
        score_thresh: Float between [0, 1] for object detector.
        nms_thresh: Float between [0, 1] indicating the non-maximum supression.
        offsets: List of length two containing floats e.g. (x_scale, y_scale)
        epsilon: Float. Values below this value would be replaced by 0.
        draw: Boolean. If ``True`` prediction are drawn in the returned image.

    # Returns
        Dictionary with inferred boxes2D, poses6D and image.
    """
    def __init__(self, camera, score_thresh=0.50, nms_thresh=0.45,
                 offsets=[0.5, 0.5], epsilon=0.15, resize=False, draw=True):
        self.detect = SSD300FAT(score_thresh, nms_thresh, draw=False)
        self.estimate_pose = MultiPowerDrillPIX2POSE6D(
            camera, offsets, epsilon, resize, draw)

    def call(self, image):
        return self.estimate_pose(image, self.detect(image)['boxes2D'])


class MultiInstanceMultiClassPIX2POSE6D(Processor):
    """Predicts poses6D of multiple instances of multiple objects from an image

    # Arguments
        detect: Function that takes as input an image and outputs a dictionary
            containing Boxes2D messages.
        name_to_model: Dictionary with class name as key and as value a
            Keras segmentation model.
        name_to_size: Dictionary with class name as key and as value the
            object sizes.
        camera: PAZ Camera with intrinsic matrix.
        offsets: List of length two containing floats e.g. (x_scale, y_scale)
        epsilon: Float. Values below this value would be replaced by 0.
        resize: Boolean. If True RGB mask is resized before computing PnP.
        draw: Boolean. If True drawing functions are applied to output image.

    # Returns
        Dictionary with inferred boxes2D, poses6D and image.
    """
    def __init__(self, detect, name_to_model, name_to_size, camera, offsets,
                 epsilon=0.15, resize=False, draw=True):
        super(MultiInstanceMultiClassPIX2POSE6D, self).__init__()
        if set(name_to_model.keys()) != set(name_to_size.keys()):
            raise ValueError('models and sizes must have same class names')
        self.detect = detect
        self.name_to_pix2points = self._build_pix2points(
            name_to_model, name_to_size, epsilon, resize)
        valid_names = list(self.name_to_model.keys())
        self.postprocess_boxes = PostprocessBoxes2D(offsets, valid_names)
        self.draw_boxes2D = pr.DrawBoxes2D(valid_names)
        self.draw_RGBmask = self._build_draw_RGBmask(name_to_size)
        self.draw_pose6D = self._build_draw_pose6D(name_to_size, camera)
        self.wrap = pr.WrapOutput(['image', 'boxes2D', 'points3D', 'poses6D'])
        self.solvePnP = pr.SolveChangingObjectPnPRANSAC(camera.intrinsics)
        self.clip = pr.ClipBoxes2D()
        self.crop = pr.CropBoxes2D()
        self.draw = draw

    def _build_pix2points(self, name_to_model, name_to_size, epsilon, resize):
        name_to_pix2points = {}
        print(name_to_model)
        for name, model in name_to_model.items():
            pix2points = Pix2Points(model, name_to_size[name], epsilon, resize)
            name_to_pix2points[name] = pix2points
        return name_to_pix2points

    def _build_draw_pose6D(self, name_to_size, camera):
        name_to_draw = {}
        for name, object_sizes in name_to_size.items():
            draw = pr.DrawPose6D(object_sizes, camera.intrinsics)
            name_to_draw[name] = draw
        return name_to_draw

    def _build_draw_RGBmask(self, name_to_size):
        name_to_draw = {}
        for name, object_sizes in name_to_size.items():
            draw = pr.DrawRGBMask(object_sizes)
            name_to_draw[name] = draw
        return name_to_draw

    def estimate_pose(self, image, box2D):
        inferences = self.name_to_pix2points[box2D.class_name](image)
        points2D = inferences['points2D']
        points3D = inferences['points3D']
        points2D = denormalize_keypoints2D(points2D, *image.shape[:2])
        points2D = translate_points2D_origin(points2D, box2D.coordinates)
        pose6D = None
        if len(points3D) > self.solvePnP.MIN_REQUIRED_POINTS:
            success, R, T = self.solvePnP(points3D, points2D)
            if success:
                pose6D = Pose6D.from_rotation_vector(R, T, box2D.class_name)
        return points2D, points3D, pose6D

    def call(self, image):
        boxes2D = self.detect(image)['boxes2D']
        boxes2D = self.postprocess_boxes(boxes2D)
        boxes2D = self.clip(image, boxes2D)
        cropped_images = self.crop(image, boxes2D)
        points2D, points3D, poses6D = [], [], []
        for crop, box2D in zip(cropped_images, boxes2D):
            inferences = self.estimate_pose(crop, box2D)
            append_lists(inferences, [points2D, points3D, poses6D])
        if self.draw:
            image = self.draw_boxes2D(image, boxes2D)
            for box2D, pose6D in zip(boxes2D, poses6D):
                name = box2D.class_name
                image = self.draw_pose6D[name](image, pose6D)
            for box2D, p2D, p3D in zip(boxes2D, points2D, points3D):
                image = self.draw_RGBmask[name](image, p2D, p3D)
        return self.wrap(image, boxes2D, points3D, poses6D)


class PIX2YCBTools6D(MultiInstanceMultiClassPIX2POSE6D):
    """Predicts poses6D of multiple instances of the YCB tools:
        '035_power_drill', '051_large_clamp', '037_scissors'

    # Arguments
        camera: PAZ Camera with intrinsic matrix.
        score_thresh: Float between [0, 1] for filtering Boxes2D.
        nsm_thresh: Float between [0, 1] non-maximum-supression filtering.
        offsets: List of length two containing floats e.g. (x_scale, y_scale)
        epsilon: Float. Values below this value would be replaced by 0.
        resize: Boolean. If True RGB mask is resized before computing PnP.
        draw: Boolean. If True drawing functions are applied to output image.

    # Returns
        Dictionary with inferred boxes2D, poses6D and image.
    """
    def __init__(self, camera, score_thresh=0.45, nms_thresh=0.15,
                 offsets=[0.25, 0.25], epsilon=0.15, resize=False, draw=True):

        self.detect = SSD300FAT(score_thresh, nms_thresh, draw=False)
        self.name_to_sizes = self._build_name_to_sizes()
        self.name_to_model = self._build_name_to_model()
        super(PIX2YCBTools6D, self).__init__(
            self.detect, self.name_to_model, self.name_to_sizes, camera,
            offsets, epsilon, resize, draw)

    def _build_name_to_model(self):
        URL = ('https://github.com/oarriaga/altamira-data/'
               'releases/download/v0.13/')

        UNET_power_drill = UNET_VGG16(3, (128, 128, 3))
        name = 'UNET-VGG16_POWERDRILL_weights.hdf5'
        weights_path = get_file(name, URL + name, cache_subdir='paz/models')
        UNET_power_drill.load_weights(weights_path)

        UNET_large_clamp = UNET_VGG16(3, (128, 128, 3))
        name = 'UNET-VGG16_LARGE-CLAMP_weights.hdf5'
        weights_path = get_file(name, URL + name, cache_subdir='paz/models')
        UNET_large_clamp.load_weights(weights_path)

        UNET_scissors = UNET_VGG16(3, (128, 128, 3))
        name = 'UNET-VGG16_SCISSORS_weights.hdf5'
        weights_path = get_file(name, URL + name, cache_subdir='paz/models')
        UNET_scissors.load_weights(weights_path)

        name_to_model = {'035_power_drill': UNET_power_drill,
                         '051_large_clamp': UNET_large_clamp,
                         '037_scissors': UNET_scissors
                         }
        return name_to_model

    def _build_name_to_sizes(self):
        name_to_sizes = {
            '035_power_drill': np.array([1840, 1874, 572]) / 10000,
            '051_large_clamp': np.array([2022, 1652, 362]) / 10000,
            '037_scissors': np.array([960, 2014, 156]) / 10000
        }
        return name_to_sizes



from .focal_loss import FocalLoss
from .focal_loss import compute_focal_loss

from .dice_loss import DiceLoss
from .dice_loss import compute_F_beta_score

from .jaccard_loss import JaccardLoss
from .jaccard_loss import compute_jaccard_score

from .weighted_reconstruction import WeightedReconstruction
from .weighted_reconstruction import WeightedReconstructionWithError
from .weighted_reconstruction import (
    compute_weighted_reconstruction_loss,
    compute_weighted_reconstruction_loss_with_error)



from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import Activation
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.layers import Flatten
from tensorflow.keras.layers import Reshape
from tensorflow.keras.layers import Concatenate
from tensorflow.keras.regularizers import l2
import tensorflow.keras.backend as K

from ..layers import Conv2DNormalization

import numpy as np
from itertools import product


def create_multibox_head(tensors, num_classes, num_priors, l2_loss=0.0005,
                         num_regressions=4, l2_norm=False, batch_norm=False):
    """Adds multibox head with classification and regression output tensors.

    # Arguments
        tensors: List of tensors.
        num_classes: Int. Number of classes.
        num_priors. List of integers. Length should equal to tensors length.
            Each integer represents the amount of bounding boxes shapes in
            each feature map value.
        l2_loss: Float. L2 loss value to be added to convolutional layers.
        num_regressions: Number of values to be regressed per prior box.
            e.g. for 2D bounding boxes we regress 4 coordinates.
        l2_norm: Boolean. If `True` l2 normalization layer is applied to
            each before a convolutional layer.
        batch_norm: Boolean. If `True` batch normalization is applied after
            each convolutional layer.
    """
    classification_layers, regression_layers = [], []
    for layer_arg, base_layer in enumerate(tensors):
        if l2_norm:
            base_layer = Conv2DNormalization(20)(base_layer)

        # classification leaf -------------------------------------------------
        num_kernels = num_priors[layer_arg] * num_classes
        class_leaf = Conv2D(num_kernels, 3, padding='same',
                            kernel_regularizer=l2(l2_loss))(base_layer)
        if batch_norm:
            class_leaf = BatchNormalization()(class_leaf)
        class_leaf = Flatten()(class_leaf)
        classification_layers.append(class_leaf)

        # regression leaf -----------------------------------------------------
        num_kernels = num_priors[layer_arg] * num_regressions
        regress_leaf = Conv2D(num_kernels, 3, padding='same',
                              kernel_regularizer=l2(l2_loss))(base_layer)
        if batch_norm:
            regress_leaf = BatchNormalization()(regress_leaf)

        regress_leaf = Flatten()(regress_leaf)
        regression_layers.append(regress_leaf)

    classifications = Concatenate(axis=1)(classification_layers)
    regressions = Concatenate(axis=1)(regression_layers)
    num_boxes = K.int_shape(regressions)[-1] // num_regressions
    classifications = Reshape((num_boxes, num_classes))(classifications)
    classifications = Activation('softmax')(classifications)
    regressions = Reshape((num_boxes, num_regressions))(regressions)
    outputs = Concatenate(
        axis=2, name='boxes')([regressions, classifications])
    return outputs


def create_prior_boxes(configuration_name='VOC'):
    configuration = get_prior_box_configuration(configuration_name)
    image_size = configuration['image_size']
    feature_map_sizes = configuration['feature_map_sizes']
    min_sizes = configuration['min_sizes']
    max_sizes = configuration['max_sizes']
    steps = configuration['steps']
    model_aspect_ratios = configuration['aspect_ratios']
    mean = []
    for feature_map_arg, feature_map_size in enumerate(feature_map_sizes):
        step = steps[feature_map_arg]
        min_size = min_sizes[feature_map_arg]
        max_size = max_sizes[feature_map_arg]
        aspect_ratios = model_aspect_ratios[feature_map_arg]
        for y, x in product(range(feature_map_size), repeat=2):
            f_k = image_size / step
            center_x = (x + 0.5) / f_k
            center_y = (y + 0.5) / f_k
            s_k = min_size / image_size
            mean = mean + [center_x, center_y, s_k, s_k]
            s_k_prime = np.sqrt(s_k * (max_size / image_size))
            mean = mean + [center_x, center_y, s_k_prime, s_k_prime]
            for aspect_ratio in aspect_ratios:
                mean = mean + [center_x, center_y, s_k * np.sqrt(aspect_ratio),
                               s_k / np.sqrt(aspect_ratio)]
                mean = mean + [center_x, center_y, s_k / np.sqrt(aspect_ratio),
                               s_k * np.sqrt(aspect_ratio)]

    output = np.asarray(mean).reshape((-1, 4))
    # output = np.clip(output, 0, 1)
    return output


def get_prior_box_configuration(configuration_name='VOC'):
    if configuration_name in {'VOC', 'FAT'}:
        configuration = {
            'feature_map_sizes': [38, 19, 10, 5, 3, 1],
            'image_size': 300,
            'steps': [8, 16, 32, 64, 100, 300],
            'min_sizes': [30, 60, 111, 162, 213, 264],
            'max_sizes': [60, 111, 162, 213, 264, 315],
            'aspect_ratios': [[2], [2, 3], [2, 3], [2, 3], [2], [2]],
            'variance': [0.1, 0.2]}

    elif configuration_name in {'COCO', 'YCBVideo'}:
        configuration = {
            'feature_map_sizes': [64, 32, 16, 8, 4, 2, 1],
            'image_size': 512,
            'steps': [8, 16, 32, 64, 128, 256, 512],
            'min_sizes': [21, 51, 133, 215, 297, 379, 461],
            'max_sizes': [51, 133, 215, 297, 379, 461, 542],
            'aspect_ratios': [[2], [2, 3], [2, 3],
                              [2, 3], [2, 3], [2], [2]],
            'variance': [0.1, 0.2]}
    else:
        raise ValueError('Invalid configuration name:', configuration_name)
    return configuration



from .image import AugmentImage
from .image import PreprocessImage
from .image import AutoEncoderPredictor
from .image import EncoderPredictor
from .image import DecoderPredictor
from .image import PreprocessImageHigherHRNet

from .detection import AugmentBoxes
from .detection import PreprocessBoxes
from .detection import AugmentDetection
from .detection import PostprocessBoxes2D
from .detection import DetectSingleShot
from .detection import SSD512COCO
from .detection import SSD512YCBVideo
from .detection import SSD300VOC
from .detection import SSD300FAT
from .detection import DetectHaarCascade
from .detection import HaarCascadeFrontalFace
from .detection import DetectMiniXceptionFER
from .detection import DetectKeypoints2D
from .detection import DetectFaceKeypointNet2D32
from .detection import SSD512HandDetection
from .detection import SSD512MinimalHandPose
from .detection import SSDPreprocess
from .detection import SSDPostprocess
from .detection import DetectSingleShotEfficientDet
from .detection import EfficientDetPreprocess
from .detection import EfficientDetPostprocess
from .detection import EFFICIENTDETD0COCO
from .detection import EFFICIENTDETD1COCO
from .detection import EFFICIENTDETD2COCO
from .detection import EFFICIENTDETD3COCO
from .detection import EFFICIENTDETD4COCO
from .detection import EFFICIENTDETD5COCO
from .detection import EFFICIENTDETD6COCO
from .detection import EFFICIENTDETD7COCO
from .detection import EFFICIENTDETD0VOC

from .keypoints import KeypointNetSharedAugmentation
from .keypoints import KeypointNetInference
from .keypoints import EstimateKeypoints2D
from .keypoints import FaceKeypointNet2D32
from .keypoints import GetKeypoints
from .keypoints import TransformKeypoints
from .keypoints import HigherHRNetHumanPose2D
from .keypoints import DetNetHandKeypoints
from .keypoints import MinimalHandPoseEstimation
from .keypoints import DetectMinimalHand
from .keypoints import EstimateHumanPose3D
from .keypoints import EstimateHumanPose

from .renderer import RenderTwoViews
from .renderer import RandomizeRenderedImage

from .classification import MiniXceptionFER
from .classification import ClassifyHandClosure

from .pose import EstimatePoseKeypoints
from .pose import HeadPoseKeypointNet2D32
from .pose import SingleInstancePIX2POSE6D
from .pose import MultiInstancePIX2POSE6D
from .pose import MultiInstanceMultiClassPIX2POSE6D
from .pose import SinglePowerDrillPIX2POSE6D
from .pose import MultiPowerDrillPIX2POSE6D
from .pose import PIX2POSEPowerDrill
from .pose import PIX2YCBTools6D

from .masks import RGBMaskToImagePoints2D
from .masks import RGBMaskToObjectPoints3D
from .masks import PredictRGBMask
from .masks import Pix2Points

from .heatmaps import GetHeatmapsAndTags

from .angles import IKNetHandJointAngles



import os
import json
import glob
from pathlib import Path
from datetime import datetime


def build_directory(root='experiments', label=None):
    """Builds and makes directory with time date and user given label.

    # Arguments:
        root: String with partial or full path.
        label: String user label.

    # Returns
        Full directory path
    """
    directory_name = build_directory_name(root, label)
    make_directory(directory_name)
    return directory_name


def build_directory_name(root, label=None):
    """Build directory name with time date and user label

    # Arguments:
        root: String with partial or full path.
        label: String user label.

    # Returns
        Full directory path
    """
    directory_name = [datetime.now().strftime("%d-%m-%Y_%H-%M-%S")]
    if label is not None:
        directory_name.extend([label])
    directory_name = '_'.join(directory_name)
    return os.path.join(root, directory_name)


def make_directory(directory_name):
    """Makes directory.

    # Arguments:
        directory_name: String. Directory name.
    """
    Path(directory_name).mkdir(parents=True, exist_ok=True)


def write_dictionary(dictionary, directory, filename, indent=4):
    """Writes dictionary as json file.

    # Arguments:
        dictionary: Dictionary to write in memory.
        directory: String. Directory name.
        filename: String. Filename.
        indent: Number of spaces between keys.
    """
    fielpath = os.path.join(directory, filename)
    filedata = open(fielpath, 'w')
    json.dump(dictionary, filedata, indent=indent)


def write_weights(model, directory, name=None):
    """Writes Keras weights in memory.

    # Arguments:
        model: Keras model.
        directory: String. Directory name.
        name: String or `None`. Weights filename.
    """
    name = model.name if name is None else name
    weights_path = os.path.join(directory, name + '_weights.hdf5')
    model.save_weights(weights_path)


def find_path(wildcard):
    filenames = glob.glob(wildcard)
    filepaths = []
    for filename in filenames:
        if os.path.isdir(filename):
            filepaths.append(filename)
    return max(filepaths, key=os.path.getmtime)


def load_latest(wildcard, filename):
    filepath = find_path(wildcard)
    filepath = os.path.join(filepath, filename)
    filedata = open(filepath, 'r')
    parameters = json.load(filedata)
    return parameters



import numpy as np

from .. import processors as pr
from ..abstract import SequentialProcessor, Processor
from ..models import (
    SSD512, SSD300, HaarCascadeDetector, EFFICIENTDETD0, EFFICIENTDETD1,
    EFFICIENTDETD2, EFFICIENTDETD3, EFFICIENTDETD4, EFFICIENTDETD5,
    EFFICIENTDETD6, EFFICIENTDETD7)
from ..datasets import get_class_names

from .image import AugmentImage, PreprocessImage
from .classification import MiniXceptionFER
from .keypoints import FaceKeypointNet2D32, DetectMinimalHand
from .keypoints import MinimalHandPoseEstimation
from ..backend.boxes import change_box_coordinates


class AugmentBoxes(SequentialProcessor):
    """Perform data augmentation with bounding boxes.

    # Arguments
        mean: List of three elements used to fill empty image spaces.
    """
    def __init__(self, mean=pr.BGR_IMAGENET_MEAN):
        super(AugmentBoxes, self).__init__()
        self.add(pr.ToImageBoxCoordinates())
        self.add(pr.Expand(mean=mean))
        self.add(pr.RandomSampleCrop())
        self.add(pr.RandomFlipBoxesLeftRight())
        self.add(pr.ToNormalizedBoxCoordinates())


class PreprocessBoxes(SequentialProcessor):
    """Preprocess bounding boxes

    # Arguments
        num_classes: Int.
        prior_boxes: Numpy array of shape ``[num_boxes, 4]`` containing
            prior/default bounding boxes.
        IOU: Float. Intersection over union used to match boxes.
        variances: List of two floats indicating variances to be encoded
            for encoding bounding boxes.
    """
    def __init__(self, num_classes, prior_boxes, IOU, variances):
        super(PreprocessBoxes, self).__init__()
        self.add(pr.MatchBoxes(prior_boxes, IOU),)
        self.add(pr.EncodeBoxes(prior_boxes, variances))
        self.add(pr.BoxClassToOneHotVector(num_classes))


class AugmentDetection(SequentialProcessor):
    """Augment boxes and images for object detection.

    # Arguments
        prior_boxes: Numpy array of shape ``[num_boxes, 4]`` containing
            prior/default bounding boxes.
        split: Flag from `paz.processors.TRAIN`, ``paz.processors.VAL``
            or ``paz.processors.TEST``. Certain transformations would take
            place depending on the flag.
        num_classes: Int.
        size: Int. Image size.
        mean: List of three elements indicating the per channel mean.
        IOU: Float. Intersection over union used to match boxes.
        variances: List of two floats indicating variances to be encoded
            for encoding bounding boxes.
    """
    def __init__(self, prior_boxes, split=pr.TRAIN, num_classes=21, size=300,
                 mean=pr.BGR_IMAGENET_MEAN, IOU=.5,
                 variances=[0.1, 0.1, 0.2, 0.2]):
        super(AugmentDetection, self).__init__()
        # image processors
        self.augment_image = AugmentImage()
        # self.augment_image.add(pr.ConvertColorSpace(pr.RGB2BGR))
        self.preprocess_image = PreprocessImage((size, size), mean)
        self.preprocess_image.insert(0, pr.ConvertColorSpace(pr.RGB2BGR))

        # box processors
        self.augment_boxes = AugmentBoxes()
        args = (num_classes, prior_boxes, IOU, variances)
        self.preprocess_boxes = PreprocessBoxes(*args)

        # pipeline
        self.add(pr.UnpackDictionary(['image', 'boxes']))
        self.add(pr.ControlMap(pr.LoadImage(), [0], [0]))
        if split == pr.TRAIN:
            self.add(pr.ControlMap(self.augment_image, [0], [0]))
            self.add(pr.ControlMap(self.augment_boxes, [0, 1], [0, 1]))
        self.add(pr.ControlMap(self.preprocess_image, [0], [0]))
        self.add(pr.ControlMap(self.preprocess_boxes, [1], [1]))
        self.add(pr.SequenceWrapper(
            {0: {'image': [size, size, 3]}},
            {1: {'boxes': [len(prior_boxes), 4 + num_classes]}}))


class PostprocessBoxes2D(SequentialProcessor):
    """Filters, squares and offsets 2D bounding boxes

    # Arguments
        valid_names: List of strings containing class names to keep.
        offsets: List of length two containing floats e.g. (x_scale, y_scale)
    """
    def __init__(self, offsets, valid_names=None):
        super(PostprocessBoxes2D, self).__init__()
        if valid_names is not None:
            self.add(pr.FilterClassBoxes2D(valid_names))
        self.add(pr.SquareBoxes2D())
        self.add(pr.OffsetBoxes2D(offsets))


class DetectSingleShot(Processor):
    """Single-shot object detection prediction.

    # Arguments
        model: Keras model.
        class_names: List of strings indicating the class names.
        preprocess: Callable, pre-processing pipeline.
        postprocess: Callable, post-processing pipeline.
        score_thresh: Float between [0, 1]
        nms_thresh: Float between [0, 1].
        variances: List, of floats.
        draw: Boolean. If ``True`` prediction are drawn in the
            returned image.
    """
    def __init__(self, model, class_names, score_thresh, nms_thresh,
                 preprocess=None, postprocess=None,
                 variances=[0.1, 0.1, 0.2, 0.2], draw=True):
        self.model = model
        self.class_names = class_names
        self.score_thresh = score_thresh
        self.nms_thresh = nms_thresh
        self.variances = variances
        self.draw = draw
        if preprocess is None:
            preprocess = SSDPreprocess(model)
        if postprocess is None:
            postprocess = SSDPostprocess(
                model, class_names, score_thresh, nms_thresh)

        super(DetectSingleShot, self).__init__()
        self.predict = pr.Predict(self.model, preprocess, postprocess)
        self.denormalize = pr.DenormalizeBoxes2D()
        self.draw_boxes2D = pr.DrawBoxes2D(self.class_names)
        self.wrap = pr.WrapOutput(['image', 'boxes2D'])

    def call(self, image):
        boxes2D = self.predict(image)
        boxes2D = self.denormalize(image, boxes2D)
        if self.draw:
            image = self.draw_boxes2D(image, boxes2D)
        return self.wrap(image, boxes2D)


class SSDPreprocess(SequentialProcessor):
    """Preprocessing pipeline for SSD.

    # Arguments
        model: Keras model.
        mean: List, of three elements indicating the per channel mean.
        color_space: Int, specifying the color space to transform.
    """
    def __init__(
            self, model, mean=pr.BGR_IMAGENET_MEAN, color_space=pr.RGB2BGR):
        super(SSDPreprocess, self).__init__()
        self.add(pr.ResizeImage(model.input_shape[1:3]))
        self.add(pr.ConvertColorSpace(color_space))
        self.add(pr.SubtractMeanImage(mean))
        self.add(pr.CastImage(float))
        self.add(pr.ExpandDims(axis=0))


class SSDPostprocess(SequentialProcessor):
    """Postprocessing pipeline for SSD.

    # Arguments
        model: Keras model.
        class_names: List, of strings indicating the class names.
        score_thresh: Float, between [0, 1]
        nms_thresh: Float, between [0, 1].
        variances: List, of floats.
        class_arg: Int, index of class to be removed.
        box_method: Int, type of boxes to boxes2D conversion method.
    """
    def __init__(self, model, class_names, score_thresh, nms_thresh,
                 variances=[0.1, 0.1, 0.2, 0.2], class_arg=0, box_method=0):
        super(SSDPostprocess, self).__init__()
        self.add(pr.Squeeze(axis=None))
        self.add(pr.DecodeBoxes(model.prior_boxes, variances))
        self.add(pr.RemoveClass(class_names, class_arg, renormalize=False))
        self.add(pr.NonMaximumSuppressionPerClass(nms_thresh))
        self.add(pr.MergeNMSBoxWithClass())
        self.add(pr.FilterBoxes(class_names, score_thresh))
        self.add(pr.ToBoxes2D(class_names, box_method))


class DetectSingleShotEfficientDet(Processor):
    """Single-shot object detection prediction for EfficientDet models.

    # Arguments
        model: Keras model.
        class_names: List of strings indicating class names.
        preprocess: Callable, preprocessing pipeline.
        postprocess: Callable, postprocessing pipeline.
        draw: Bool. If ``True`` prediction are drawn on the
            returned image.

    # Properties
        model: Keras model.
        draw: Bool.
        preprocess: Callable.
        postprocess: Callable.
        draw_boxes2D: Callable.
        wrap: Callable.

    # Methods
        call()
    """
    def __init__(self, model, class_names, score_thresh, nms_thresh,
                 preprocess=None, postprocess=None, draw=True):
        self.model = model
        self.draw = draw
        self.draw_boxes2D = pr.DrawBoxes2D(class_names)
        self.wrap = pr.WrapOutput(['image', 'boxes2D'])
        if preprocess is None:
            self.preprocess = EfficientDetPreprocess(model)
        if postprocess is None:
            self.postprocess = EfficientDetPostprocess(
                model, class_names, score_thresh, nms_thresh)
        super(DetectSingleShotEfficientDet, self).__init__()

    def call(self, image):
        preprocessed_image, image_scales = self.preprocess(image)
        outputs = self.model(preprocessed_image)
        outputs = change_box_coordinates(outputs)
        boxes2D = self.postprocess(outputs, image_scales)
        if self.draw:
            image = self.draw_boxes2D(image, boxes2D)
        return self.wrap(image, boxes2D)


class EfficientDetPreprocess(SequentialProcessor):
    """Preprocessing pipeline for EfficientDet.

    # Arguments
        model: Keras model.
        mean: Tuple, containing mean per channel on ImageNet.
        standard_deviation: Tuple, containing standard deviations
            per channel on ImageNet.
    """
    def __init__(self, model, mean=pr.RGB_IMAGENET_MEAN,
                 standard_deviation=pr.RGB_IMAGENET_STDEV):
        super(EfficientDetPreprocess, self).__init__()
        self.add(pr.CastImage(float))
        self.add(pr.SubtractMeanImage(mean=mean))
        self.add(pr.DivideStandardDeviationImage(standard_deviation))
        self.add(pr.ScaledResize(image_size=model.input_shape[1]))


class EfficientDetPostprocess(Processor):
    """Postprocessing pipeline for EfficientDet.

    # Arguments
        model: Keras model.
        class_names: List of strings indicating class names.
        score_thresh: Float between [0, 1].
        nms_thresh: Float between [0, 1].
        variances: List of float values.
        class_arg: Int, index of the class to be removed.
        renormalize: Bool, if true scores are renormalized.
        method: Int, method to convert boxes to ``Boxes2D``.
    """
    def __init__(self, model, class_names, score_thresh, nms_thresh,
                 variances=[1.0, 1.0, 1.0, 1.0], class_arg=None):
        super(EfficientDetPostprocess, self).__init__()
        model.prior_boxes = model.prior_boxes * model.input_shape[1]
        self.postprocess = pr.SequentialProcessor([
            pr.Squeeze(axis=None),
            pr.DecodeBoxes(model.prior_boxes, variances),
            pr.RemoveClass(class_names, class_arg)])
        self.scale = pr.ScaleBox()
        self.nms_per_class = pr.NonMaximumSuppressionPerClass(nms_thresh)
        self.merge_box_and_class = pr.MergeNMSBoxWithClass()
        self.filter_boxes = pr.FilterBoxes(class_names, score_thresh)
        self.to_boxes2D = pr.ToBoxes2D(class_names)
        self.round_boxes = pr.RoundBoxes2D()

    def call(self, output, image_scale):
        box_data = self.postprocess(output)
        box_data = self.scale(box_data, image_scale)
        box_data, class_labels = self.nms_per_class(box_data)
        box_data = self.merge_box_and_class(box_data, class_labels)
        box_data = self.filter_boxes(box_data)
        boxes2D = self.to_boxes2D(box_data)
        boxes2D = self.round_boxes(boxes2D)
        return boxes2D


class SSD512COCO(DetectSingleShot):
    """Single-shot inference pipeline with SSD512 trained on COCO.

    # Arguments
        score_thresh: Float between [0, 1]
        nms_thresh: Float between [0, 1].
        draw: Boolean. If ``True`` prediction are drawn in the returned image.

    # Example
        ``` python
        from paz.pipelines import SSD512COCO

        detect = SSD512COCO()

        # apply directly to an image (numpy-array)
        inferences = detect(image)
        ```
     # Returns
        A function that takes an RGB image and outputs the predictions
        as a dictionary with ``keys``: ``image`` and ``boxes2D``.
        The corresponding values of these keys contain the image with the drawn
        inferences and a list of ``paz.abstract.messages.Boxes2D``.

    # Reference
        - [SSD: Single Shot MultiBox
            Detector](https://arxiv.org/abs/1512.02325)
    """
    def __init__(self, score_thresh=0.60, nms_thresh=0.45, draw=True):
        model = SSD512()
        names = get_class_names('COCO')
        super(SSD512COCO, self).__init__(
            model, names, score_thresh, nms_thresh, draw=draw)


class SSD512YCBVideo(DetectSingleShot):
    """Single-shot inference pipeline with SSD512 trained on YCBVideo.

    # Arguments
        score_thresh: Float between [0, 1]
        nms_thresh: Float between [0, 1].
        draw: Boolean. If ``True`` prediction are drawn in the returned image.

    # Example
        ``` python
        from paz.pipelines import SSD512YCBVideo

        detect = SSD512YCBVideo()

        # apply directly to an image (numpy-array)
        inferences = detect(image)
        ```

    # Returns
        A function that takes an RGB image and outputs the predictions
        as a dictionary with ``keys``: ``image`` and ``boxes2D``.
        The corresponding values of these keys contain the image with the drawn
        inferences and a list of ``paz.abstract.messages.Boxes2D``.
    """
    def __init__(self, score_thresh=0.60, nms_thresh=0.45, draw=True):
        names = get_class_names('YCBVideo')
        model = SSD512(head_weights='YCBVideo', num_classes=len(names))
        super(SSD512YCBVideo, self).__init__(
            model, names, score_thresh, nms_thresh, draw=draw)


class SSD300VOC(DetectSingleShot):
    """Single-shot inference pipeline with SSD300 trained on VOC.

    # Arguments
        score_thresh: Float between [0, 1]
        nms_thresh: Float between [0, 1].
        draw: Boolean. If ``True`` prediction are drawn in the returned image.

    # Example
        ``` python
        from paz.pipelines import SSD300VOC

        detect = SSD300VOC()

        # apply directly to an image (numpy-array)
        inferences = detect(image)
        ```

    # Returns
        A function that takes an RGB image and outputs the predictions
        as a dictionary with ``keys``: ``image`` and ``boxes2D``.
        The corresponding values of these keys contain the image with the drawn
        inferences and a list of ``paz.abstract.messages.Boxes2D``.

    # Reference
        - [SSD: Single Shot MultiBox
            Detector](https://arxiv.org/abs/1512.02325)
    """
    def __init__(self, score_thresh=0.60, nms_thresh=0.45, draw=True):
        model = SSD300()
        names = get_class_names('VOC')
        super(SSD300VOC, self).__init__(
            model, names, score_thresh, nms_thresh, draw=draw)


class SSD300FAT(DetectSingleShot):
    """Single-shot inference pipeline with SSD300 trained on FAT.

    # Arguments
        score_thresh: Float between [0, 1]
        nms_thresh: Float between [0, 1].
        draw: Boolean. If ``True`` prediction are drawn in the returned image.

    # Example
        ``` python
        from paz.pipelines import SSD300FAT

        detect = SSD300FAT()

        # apply directly to an image (numpy-array)
        inferences = detect(image)
        ```
    # Returns
        A function that takes an RGB image and outputs the predictions
        as a dictionary with ``keys``: ``image`` and ``boxes2D``.
        The corresponding values of these keys contain the image with the drawn
        inferences and a list of ``paz.abstract.messages.Boxes2D``.

    """
    def __init__(self, score_thresh=0.60, nms_thresh=0.45, draw=True):
        model = SSD300(22, 'FAT', 'FAT')
        names = get_class_names('FAT')
        super(SSD300FAT, self).__init__(
            model, names, score_thresh, nms_thresh, draw=draw)


class DetectHaarCascade(Processor):
    """HaarCascade prediction pipeline/function from RGB-image.

    # Arguments
        detector: An instantiated ``HaarCascadeDetector`` model.
        offsets: List of two elements. Each element must be between [0, 1].
        class_names: List of strings.
        draw: Boolean. If ``True`` prediction are drawn in the returned image.

    # Returns
        A function for predicting bounding box detections.
    """
    def __init__(self, detector, class_names=None, colors=None, draw=True):
        super(DetectHaarCascade, self).__init__()
        self.detector = detector
        self.class_names = class_names
        self.colors = colors
        self.draw = draw
        RGB2GRAY = pr.ConvertColorSpace(pr.RGB2GRAY)
        postprocess = SequentialProcessor()
        postprocess.add(pr.ToBoxes2D(self.class_names, box_method=2))
        self.predict = pr.Predict(self.detector, RGB2GRAY, postprocess)
        self.draw_boxes2D = pr.DrawBoxes2D(self.class_names, self.colors)
        self.wrap = pr.WrapOutput(['image', 'boxes2D'])

    def call(self, image):
        boxes2D = self.predict(image)
        if self.draw:
            image = self.draw_boxes2D(image, boxes2D)
        return self.wrap(image, boxes2D)


class HaarCascadeFrontalFace(DetectHaarCascade):
    """HaarCascade pipeline for detecting frontal faces

    # Arguments
        class_name: String indicating the class name.
        color: List indicating the RGB color e.g. ``[0, 255, 0]``.
        draw: Boolean. If ``False`` the bounding boxes are not drawn.

    # Example
        ``` python
        from paz.pipelines import HaarCascadeFrontalFace

        detect = HaarCascadeFrontalFace()

        # apply directly to an image (numpy-array)
        inferences = detect(image)
        ```
    # Returns
        A function that takes an RGB image and outputs the predictions
        as a dictionary with ``keys``: ``image`` and ``boxes2D``.
        The corresponding values of these keys contain the image with the drawn
        inferences and a list of ``paz.abstract.messages.Boxes2D``.

    """
    def __init__(self, class_name='Face', color=[0, 255, 0], draw=True):
        self.model = HaarCascadeDetector('frontalface_default', class_arg=0)
        super(HaarCascadeFrontalFace, self).__init__(
            self.model, [class_name], [color], draw)


EMOTION_COLORS = [[255, 0, 0], [45, 90, 45], [255, 0, 255], [255, 255, 0],
                  [0, 0, 255], [0, 255, 255], [0, 255, 0]]


class DetectMiniXceptionFER(Processor):
    """Emotion classification and detection pipeline.

    # Returns
        Dictionary with ``image`` and ``boxes2D``.

    # Example
        ``` python
        from paz.pipelines import DetectMiniXceptionFER

        detect = DetectMiniXceptionFER()

        # apply directly to an image (numpy-array)
        inferences = detect(image)
        ```
    # Returns
        A function that takes an RGB image and outputs the predictions
        as a dictionary with ``keys``: ``image`` and ``boxes2D``.
        The corresponding values of these keys contain the image with the drawn
        inferences and a list of ``paz.abstract.messages.Boxes2D``.

    # References
       - [Real-time Convolutional Neural Networks for Emotion and
            Gender Classification](https://arxiv.org/abs/1710.07557)
    """
    def __init__(self, offsets=[0, 0], colors=EMOTION_COLORS):
        super(DetectMiniXceptionFER, self).__init__()
        self.offsets = offsets
        self.colors = colors

        # detection
        self.detect = HaarCascadeFrontalFace()
        self.square = SequentialProcessor()
        self.square.add(pr.SquareBoxes2D())
        self.square.add(pr.OffsetBoxes2D(offsets))
        self.clip = pr.ClipBoxes2D()
        self.crop = pr.CropBoxes2D()

        # classification
        self.classify = MiniXceptionFER()

        # drawing and wrapping
        self.class_names = self.classify.class_names
        self.draw = pr.DrawBoxes2D(self.class_names, self.colors, True)
        self.wrap = pr.WrapOutput(['image', 'boxes2D'])

    def call(self, image):
        boxes2D = self.detect(image.copy())['boxes2D']
        boxes2D = self.square(boxes2D)
        boxes2D = self.clip(image, boxes2D)
        cropped_images = self.crop(image, boxes2D)
        for cropped_image, box2D in zip(cropped_images, boxes2D):
            predictions = self.classify(cropped_image)
            box2D.class_name = predictions['class_name']
            box2D.score = np.amax(predictions['scores'])
        image = self.draw(image, boxes2D)
        return self.wrap(image, boxes2D)


class DetectKeypoints2D(Processor):
    def __init__(self, detect, estimate_keypoints, offsets=[0, 0], radius=3):
        """General detection and keypoint estimator pipeline.

        # Arguments
            detect: Function for detecting objects. The output should be a
                dictionary with key ``Boxes2D`` containing a list
                of ``Boxes2D`` messages.
            estimate_keypoints: Function for estimating keypoints. The output
                should be a dictionary with key ``keypoints`` containing
                a numpy array of keypoints.
            offsets: List of two elements. Each element must be between [0, 1].
            radius: Int indicating the radius of the keypoints to be drawn.
        """
        super(DetectKeypoints2D, self).__init__()
        self.detect = detect
        self.estimate_keypoints = estimate_keypoints
        self.num_keypoints = estimate_keypoints.num_keypoints
        self.square = SequentialProcessor()
        self.square.add(pr.SquareBoxes2D())
        self.square.add(pr.OffsetBoxes2D(offsets))
        self.clip = pr.ClipBoxes2D()
        self.crop = pr.CropBoxes2D()
        self.change_coordinates = pr.ChangeKeypointsCoordinateSystem()
        self.draw = pr.DrawKeypoints2D(self.num_keypoints, radius, False)
        self.draw_boxes = pr.DrawBoxes2D(detect.class_names, detect.colors)
        self.wrap = pr.WrapOutput(['image', 'boxes2D', 'keypoints'])

    def call(self, image):
        boxes2D = self.detect(image)['boxes2D']
        boxes2D = self.square(boxes2D)
        boxes2D = self.clip(image, boxes2D)
        cropped_images = self.crop(image, boxes2D)
        keypoints2D = []
        for cropped_image, box2D in zip(cropped_images, boxes2D):
            keypoints = self.estimate_keypoints(cropped_image)['keypoints']
            keypoints = self.change_coordinates(keypoints, box2D)
            keypoints2D.append(keypoints)
            image = self.draw(image, keypoints)
        image = self.draw_boxes(image, boxes2D)
        return self.wrap(image, boxes2D, keypoints2D)


class DetectFaceKeypointNet2D32(DetectKeypoints2D):
    """Frontal face detection pipeline with facial keypoint estimation.

    # Arguments
        offsets: List of two elements. Each element must be between [0, 1].
        radius: Int indicating the radius of the keypoints to be drawn.

    # Example
        ``` python
        from paz.pipelines import DetectFaceKeypointNet2D32

        detect = DetectFaceKeypointNet2D32()

        # apply directly to an image (numpy-array)
        inferences = detect(image)
        ```
    # Returns
        A function that takes an RGB image and outputs the predictions
        as a dictionary with ``keys``: ``image`` and ``boxes2D``.
        The corresponding values of these keys contain the image with the drawn
        inferences and a list of ``paz.abstract.messages.Boxes2D``.

    """
    def __init__(self, offsets=[0, 0], radius=3):
        detect = HaarCascadeFrontalFace(draw=False)
        estimate_keypoints = FaceKeypointNet2D32(draw=False)
        super(DetectFaceKeypointNet2D32, self).__init__(
            detect, estimate_keypoints, offsets, radius)


class SSD512HandDetection(DetectSingleShot):
    """Minimal hand detection with SSD512Custom trained on OPenImageV6.

    # Arguments
        score_thresh: Float between [0, 1]
        nms_thresh: Float between [0, 1].
        draw: Boolean. If ``True`` prediction are drawn in the returned image.

    # Example
        ``` python
        from paz.pipelines import SSD512HandDetection

        detect = SSD512HandDetection()

        # apply directly to an image (numpy-array)
        inferences = detect(image)
        ```
     # Returns
        A function that takes an RGB image and outputs the predictions
        as a dictionary with ``keys``: ``image`` and ``boxes2D``.
        The corresponding values of these keys contain the image with the drawn
        inferences and a list of ``paz.abstract.messages.Boxes2D``.

    # Reference
        - [SSD: Single Shot MultiBox
            Detector](https://arxiv.org/abs/1512.02325)
    """
    def __init__(self, score_thresh=0.40, nms_thresh=0.45, draw=True):
        class_names = ['background', 'hand']
        num_classes = len(class_names)
        model = SSD512(num_classes, base_weights='OIV6Hand',
                       head_weights='OIV6Hand')
        super(SSD512HandDetection, self).__init__(
            model, class_names, score_thresh, nms_thresh, draw=draw)


class SSD512MinimalHandPose(DetectMinimalHand):
    """Hand detection and minimal hand pose estimation pipeline.

    # Arguments
        right_hand: Boolean. True for right hand inference.
        offsets: List of two elements. Each element must be between [0, 1].

    # Example
        ``` python
        from paz.pipelines import SSD512MinimalHandPose

        detect = SSD512MinimalHandPose()

        # apply directly to an image (numpy-array)
        inferences = detect(image)
        ```

    # Returns
        A function that takes an RGB image and outputs the predictions
        as a dictionary with ``keys``: ``image``,  ``boxes2D``,
        ``Keypoints2D``, ``Keypoints3D``.
        The corresponding values of these keys contain the image with the drawn
        inferences.
    """
    def __init__(self, right_hand=False, offsets=[0.25, 0.25]):
        detector = SSD512HandDetection()
        keypoint_estimator = MinimalHandPoseEstimation(right_hand)
        super(SSD512MinimalHandPose, self).__init__(
            detector, keypoint_estimator, offsets)


class EFFICIENTDETD0COCO(DetectSingleShotEfficientDet):
    """Single-shot inference pipeline with EFFICIENTDETD0 trained
    on COCO.

    # Arguments
        score_thresh: Float between [0, 1]
        nms_thresh: Float between [0, 1].
        draw: Boolean. If ``True`` prediction are drawn in the
            returned image.

    # References
        [Google AutoML repository implementation of EfficientDet](
        https://github.com/google/automl/tree/master/efficientdet)
    """
    def __init__(self, score_thresh=0.60, nms_thresh=0.45, draw=True):
        names = get_class_names('COCO_EFFICIENTDET')
        model = EFFICIENTDETD0(num_classes=len(names),
                               base_weights='COCO', head_weights='COCO')
        super(EFFICIENTDETD0COCO, self).__init__(
            model, names, score_thresh, nms_thresh, draw=draw)


class EFFICIENTDETD1COCO(DetectSingleShotEfficientDet):
    """Single-shot inference pipeline with EFFICIENTDETD1 trained
    on COCO.

    # Arguments
        score_thresh: Float between [0, 1]
        nms_thresh: Float between [0, 1].
        draw: Boolean. If ``True`` prediction are drawn in the
            returned image.

    # References
        [Google AutoML repository implementation of EfficientDet](
        https://github.com/google/automl/tree/master/efficientdet)
    """
    def __init__(self, score_thresh=0.60, nms_thresh=0.45, draw=True):
        names = get_class_names('COCO_EFFICIENTDET')
        model = EFFICIENTDETD1(num_classes=len(names),
                               base_weights='COCO', head_weights='COCO')
        super(EFFICIENTDETD1COCO, self).__init__(
            model, names, score_thresh, nms_thresh, draw=draw)


class EFFICIENTDETD2COCO(DetectSingleShotEfficientDet):
    """Single-shot inference pipeline with EFFICIENTDETD2 trained
    on COCO.

    # Arguments
        score_thresh: Float between [0, 1]
        nms_thresh: Float between [0, 1].
        draw: Boolean. If ``True`` prediction are drawn in the
            returned image.

    # References
        [Google AutoML repository implementation of EfficientDet](
        https://github.com/google/automl/tree/master/efficientdet)
    """
    def __init__(self, score_thresh=0.60, nms_thresh=0.45, draw=True):
        names = get_class_names('COCO_EFFICIENTDET')
        model = EFFICIENTDETD2(num_classes=len(names),
                               base_weights='COCO', head_weights='COCO')
        super(EFFICIENTDETD2COCO, self).__init__(
            model, names, score_thresh, nms_thresh, draw=draw)


class EFFICIENTDETD3COCO(DetectSingleShotEfficientDet):
    """Single-shot inference pipeline with EFFICIENTDETD3 trained
    on COCO.

    # Arguments
        score_thresh: Float between [0, 1]
        nms_thresh: Float between [0, 1].
        draw: Boolean. If ``True`` prediction are drawn in the
            returned image.

    # References
        [Google AutoML repository implementation of EfficientDet](
        https://github.com/google/automl/tree/master/efficientdet)
    """
    def __init__(self, score_thresh=0.60, nms_thresh=0.45, draw=True):
        names = get_class_names('COCO_EFFICIENTDET')
        model = EFFICIENTDETD3(num_classes=len(names),
                               base_weights='COCO', head_weights='COCO')
        super(EFFICIENTDETD3COCO, self).__init__(
            model, names, score_thresh, nms_thresh, draw=draw)


class EFFICIENTDETD4COCO(DetectSingleShotEfficientDet):
    """Single-shot inference pipeline with EFFICIENTDETD4 trained
    on COCO.

    # Arguments
        score_thresh: Float between [0, 1]
        nms_thresh: Float between [0, 1].
        draw: Boolean. If ``True`` prediction are drawn in the
            returned image.

    # References
        [Google AutoML repository implementation of EfficientDet](
        https://github.com/google/automl/tree/master/efficientdet)
    """
    def __init__(self, score_thresh=0.60, nms_thresh=0.45, draw=True):
        names = get_class_names('COCO_EFFICIENTDET')
        model = EFFICIENTDETD4(num_classes=len(names),
                               base_weights='COCO', head_weights='COCO')
        super(EFFICIENTDETD4COCO, self).__init__(
            model, names, score_thresh, nms_thresh, draw=draw)


class EFFICIENTDETD5COCO(DetectSingleShotEfficientDet):
    """Single-shot inference pipeline with EFFICIENTDETD5 trained
    on COCO.

    # Arguments
        score_thresh: Float between [0, 1]
        nms_thresh: Float between [0, 1].
        draw: Boolean. If ``True`` prediction are drawn in the
            returned image.

    # References
        [Google AutoML repository implementation of EfficientDet](
        https://github.com/google/automl/tree/master/efficientdet)
    """
    def __init__(self, score_thresh=0.60, nms_thresh=0.45, draw=True):
        names = get_class_names('COCO_EFFICIENTDET')
        model = EFFICIENTDETD5(num_classes=len(names),
                               base_weights='COCO', head_weights='COCO')
        super(EFFICIENTDETD5COCO, self).__init__(
            model, names, score_thresh, nms_thresh, draw=draw)


class EFFICIENTDETD6COCO(DetectSingleShotEfficientDet):
    """Single-shot inference pipeline with EFFICIENTDETD6 trained
    on COCO.

    # Arguments
        score_thresh: Float between [0, 1]
        nms_thresh: Float between [0, 1].
        draw: Boolean. If ``True`` prediction are drawn in the
            returned image.

    # References
        [Google AutoML repository implementation of EfficientDet](
        https://github.com/google/automl/tree/master/efficientdet)
    """
    def __init__(self, score_thresh=0.60, nms_thresh=0.45, draw=True):
        names = get_class_names('COCO_EFFICIENTDET')
        model = EFFICIENTDETD6(num_classes=len(names),
                               base_weights='COCO', head_weights='COCO')
        super(EFFICIENTDETD6COCO, self).__init__(
            model, names, score_thresh, nms_thresh, draw=draw)


class EFFICIENTDETD7COCO(DetectSingleShotEfficientDet):
    """Single-shot inference pipeline with EFFICIENTDETD7 trained
    on COCO.

    # Arguments
        score_thresh: Float between [0, 1]
        nms_thresh: Float between [0, 1].
        draw: Boolean. If ``True`` prediction are drawn in the
            returned image.

    # References
        [Google AutoML repository implementation of EfficientDet](
        https://github.com/google/automl/tree/master/efficientdet)
    """
    def __init__(self, score_thresh=0.60, nms_thresh=0.45, draw=True):
        names = get_class_names('COCO_EFFICIENTDET')
        model = EFFICIENTDETD7(num_classes=len(names),
                               base_weights='COCO', head_weights='COCO')
        super(EFFICIENTDETD7COCO, self).__init__(
            model, names, score_thresh, nms_thresh, draw=draw)


class EFFICIENTDETD0VOC(DetectSingleShot):
    """Single-shot inference pipeline with EFFICIENTDETD0 trained
    on VOC.

    # Arguments
        score_thresh: Float between [0, 1]
        nms_thresh: Float between [0, 1].
        draw: Boolean. If ``True`` prediction are drawn in the
            returned image.

    # References
        [Google AutoML repository implementation of EfficientDet](
        https://github.com/google/automl/tree/master/efficientdet)
    """
    def __init__(self, score_thresh=0.60, nms_thresh=0.45, draw=True):
        names = get_class_names('VOC')
        model = EFFICIENTDETD0(num_classes=len(names),
                               base_weights='VOC', head_weights='VOC')
        super(EFFICIENTDETD0VOC, self).__init__(
            model, names, score_thresh, nms_thresh, draw=draw)



from ..abstract import SequentialProcessor
from .. import processors as pr
from ..backend.image import get_affine_transform


class AugmentImage(SequentialProcessor):
    """Augments an RGB image by randomly changing contrast, brightness
        saturation and hue.
    """
    def __init__(self):
        super(AugmentImage, self).__init__()
        self.add(pr.RandomContrast())
        self.add(pr.RandomBrightness())
        self.add(pr.RandomSaturation())
        self.add(pr.RandomHue())


class PreprocessImage(SequentialProcessor):
    """Preprocess RGB image by resizing it to the given ``shape``. If a
    ``mean`` is given it is substracted from image and it not the image gets
    normalized.

    # Arguments
        shape: List of two Ints.
        mean: List of three Ints indicating the per-channel mean to be
            subtracted.
    """
    def __init__(self, shape, mean=pr.BGR_IMAGENET_MEAN):
        super(PreprocessImage, self).__init__()
        self.add(pr.ResizeImage(shape))
        self.add(pr.CastImage(float))
        if mean is None:
            self.add(pr.NormalizeImage())
        else:
            self.add(pr.SubtractMeanImage(mean))


class AutoEncoderPredictor(SequentialProcessor):
    """Pipeline for predicting values from an auto-encoder.

    # Arguments
        model: Keras model.
    """
    def __init__(self, model):
        super(AutoEncoderPredictor, self).__init__()
        preprocess = SequentialProcessor(
            [pr.ResizeImage(model.input_shape[1:3]),
             pr.ConvertColorSpace(pr.RGB2BGR),
             pr.NormalizeImage(),
             pr.ExpandDims(0)])
        self.add(pr.Predict(model, preprocess))
        self.add(pr.Squeeze(0))
        self.add(pr.DenormalizeImage())
        self.add(pr.CastImage('uint8'))
        self.add(pr.WrapOutput(['image']))


class EncoderPredictor(SequentialProcessor):
    """Pipeline for predicting latent vector of an encoder.

    # Arguments
        model: Keras model.
    """
    def __init__(self, encoder):
        super(EncoderPredictor, self).__init__()
        self.encoder = encoder
        preprocess = SequentialProcessor([
            pr.ConvertColorSpace(pr.RGB2BGR),
            pr.ResizeImage(encoder.input_shape[1:3]),
            pr.NormalizeImage(),
            pr.ExpandDims(0)])
        self.add(pr.Predict(encoder, preprocess, pr.Squeeze(0)))


class DecoderPredictor(SequentialProcessor):
    """Pipeline for predicting decoded image from a latent vector.

    # Arguments
        model: Keras model.
    """
    def __init__(self, decoder):
        self.decoder = decoder
        super(DecoderPredictor, self).__init__()
        self.add(pr.Predict(decoder, pr.ExpandDims(0), pr.Squeeze(0)))
        self.add(pr.DenormalizeImage())
        self.add(pr.CastImage('uint8'))
        self.add(pr.ConvertColorSpace(pr.BGR2RGB))


class PreprocessImageHigherHRNet(pr.Processor):
    """Transform the image according to the HigherHRNet model requirement.
    # Arguments
        scaling_factor: Int. scale factor for image dimensions.
        input_size: Int. resize the first dimension of image to input size.
        inverse: Boolean. Reverse the affine transform input.
        image: Numpy array. Input image

    # Returns
        image: resized and transformed image
        center: center of the image
        scale: scaled image dimensions
    """
    def __init__(self, scaling_factor=200, input_size=512, multiple=64):
        super(PreprocessImageHigherHRNet, self).__init__()
        self.get_image_center = pr.GetImageCenter()
        self.get_size = pr.GetTransformationSize(input_size, multiple)
        self.get_scale = pr.GetTransformationScale(scaling_factor)
        self.get_source_destination_point = pr.GetSourceDestinationPoints(
            scaling_factor)
        self.transform_image = pr.SequentialProcessor(
            [pr.WarpAffine(), pr.ImagenetPreprocessInput(), pr.ExpandDims(0)])

    def call(self, image):
        center = self.get_image_center(image)
        size = self.get_size(image)
        scale = self.get_scale(image, size)
        source_point, destination_point = self.get_source_destination_point(
            center, scale, size)
        transform = get_affine_transform(source_point, destination_point)
        image = self.transform_image(image, transform, size)
        return image, center, scale



import tensorflow as tf
from tensorflow.keras.losses import Loss


def compute_F_beta_score(y_true, y_pred, beta=1.0, class_weights=1.0):
    """Computes the F beta score. The F beta score is the geometric mean
    of the precision and recall, where the recall is B times more important
    than the precision.

    # Arguments
        y_true: Tensor of shape ``(batch, H, W, num_channels)``.
        y_pred: Tensor of shape ``(batch, H, W, num_channels)``.
        beta: Float.
        class_weights: Float or list of floats of shape ``(num_classes)``.

    # Returns
        Tensor of shape ``(batch)`` containing the F beta score per sample.
    """
    true_positives = tf.reduce_sum(y_true * y_pred, axis=[1, 2])
    false_positives = tf.reduce_sum(y_pred, axis=[1, 2]) - true_positives
    false_negatives = tf.reduce_sum(y_true, axis=[1, 2]) - true_positives
    B_squared = tf.math.pow(beta, 2)
    numerator = (1.0 + B_squared) * true_positives
    denominator = numerator + (B_squared * false_negatives) + false_positives
    F_beta_score = numerator / (denominator + 1e-5)
    return class_weights * F_beta_score


class DiceLoss(Loss):
    """Computes the F beta loss. The F beta score is the geometric mean
    of the precision and recall, where the recall is B times more important
    than the precision.

    # Arguments
        beta: Float.
        class_weights: Float or list of floats of shape ``(num_classes)``.
    """
    def __init__(self, beta=1.0, class_weights=1.0):
        super(DiceLoss, self).__init__()
        self.beta = beta
        self.class_weights = class_weights

    def call(self, y_true, y_pred):
        args = (self.beta, self.class_weights)
        return 1.0 - compute_F_beta_score(y_true, y_pred, *args)



from .detection import SSD300
from .detection import SSD512
from .detection import HaarCascadeDetector
from .detection import EFFICIENTDETD0
from .detection import EFFICIENTDETD1
from .detection import EFFICIENTDETD2
from .detection import EFFICIENTDETD3
from .detection import EFFICIENTDETD4
from .detection import EFFICIENTDETD5
from .detection import EFFICIENTDETD6
from .detection import EFFICIENTDETD7
from .keypoint.simplebaselines import SimpleBaseline
from .keypoint.projector import Projector
from .keypoint.keypointnet import KeypointNet
from .keypoint.keypointnet import KeypointNetShared
from .keypoint.keypointnet import KeypointNet2D
from .keypoint.hrnet import HRNetResidual
from .keypoint.hrnet import HRNetDense
from .keypoint.detnet import DetNet
from .keypoint.iknet import IKNet
from .classification import build_xception
from .classification import MiniXception
from .classification import ProtoEmbedding
from .classification import ProtoNet
from .segmentation import UNET
from .segmentation import UNET_VGG16
from .segmentation import UNET_VGG19
from .segmentation import UNET_RESNET50
from .pose_estimation import HigherHRNet



from warnings import warn

import cv2
import numpy as np

UPNP = cv2.SOLVEPNP_UPNP
LEVENBERG_MARQUARDT = cv2.SOLVEPNP_ITERATIVE


def build_cube_points3D(width, height, depth):
    """Build the 3D points of a cube in the openCV coordinate system:
                               4--------1
                              /|       /|
                             / |      / |
                            3--------2  |
                            |  8_____|__5
                            | /      | /
                            |/       |/
                            7--------6

                   Z (depth)
                  /
                 /_____X (width)
                 |
                 |
                 Y (height)

    # Arguments
        height: float, height of the 3D box.
        width: float,  width of the 3D box.
        depth: float,  width of the 3D box.

    # Returns
        Numpy array of shape ``(8, 3)'' corresponding to 3D keypoints of a cube
    """
    half_height, half_width, half_depth = height / 2., width / 2., depth / 2.
    point_1 = [+half_width, -half_height, +half_depth]
    point_2 = [+half_width, -half_height, -half_depth]
    point_3 = [-half_width, -half_height, -half_depth]
    point_4 = [-half_width, -half_height, +half_depth]
    point_5 = [+half_width, +half_height, +half_depth]
    point_6 = [+half_width, +half_height, -half_depth]
    point_7 = [-half_width, +half_height, -half_depth]
    point_8 = [-half_width, +half_height, +half_depth]
    return np.array([point_1, point_2, point_3, point_4,
                     point_5, point_6, point_7, point_8])


def normalize_keypoints2D(points2D, height, width):
    """Transform points2D in image coordinates to normalized coordinates i.e.
        [U, V] -> [-1, 1]. UV have maximum values of [W, H] respectively.

             Image plane

                 width
           (0,0)-------->  (U)
             |
      height |
             |
             v

            (V)

    # Arguments
        points2D: Numpy array of shape (num_keypoints, 2).
        height: Int. Height of the image
        width: Int. Width of the image

    # Returns
        Numpy array of shape (num_keypoints, 2).
    """
    image_shape = np.array([width, height])
    points2D = points2D / image_shape  # [W, 0], [0, H] -> [1,  0], [0,  1]
    points2D = 2.0 * points2D          # [1, 0], [0, 1] -> [2,  0], [0,  2]
    points2D = points2D - 1.0          # [2, 0], [0, 2] -> [-1, 1], [-1, 1]
    return points2D


def denormalize_keypoints2D(points2D, height, width):
    """Transform nomralized points2D to image UV coordinates i.e.
        [-1, 1] -> [U, V]. UV have maximum values of [W, H] respectively.

             Image plane

           (0,0)-------->  (U)
             |
             |
             |
             v

            (V)

    # Arguments
        points2D: Numpy array of shape (num_keypoints, 2).
        height: Int. Height of the image
        width: Int. Width of the image

    # Returns
        Numpy array of shape (num_keypoints, 2).
    """
    image_shape = np.array([width, height])
    points2D = points2D + 1.0          # [-1, 1], [-1, 1] -> [2, 0], [0, 2]
    points2D = points2D / 2.0          # [2 , 0], [0 , 2] -> [1, 0], [0, 1]
    points2D = points2D * image_shape  # [1 , 0], [0 , 1] -> [W, 0], [0, H]
    return points2D


def cascade_classifier(path):
    """OpenCV Cascade classifier.

    # Arguments
        path: String. Path to default openCV XML format.

    # Returns
        OpenCV classifier with ``detectMultiScale`` for inference..
    """
    return cv2.CascadeClassifier(path)


def solve_PNP(points3D, points2D, camera, solver):
    """Calculates 6D pose from 3D points and 2D keypoints correspondences.

    # Arguments
        points3D: Numpy array of shape ``(num_points, 3)``.
            3D points known in advance.
        points2D: Numpy array of shape ``(num_points, 2)``.
            Predicted 2D keypoints of object.
        camera: Instance of ''paz.backend.Camera'' containing as properties
            the ''camera_intrinsics'' a Numpy array of shape ''(3, 3)''
            usually calculated from the openCV ''calibrateCamera'' function,
            and the ''distortion'' a Numpy array of shape ''(5)'' in which the
            elements are usually obtained from the openCV
            ''calibrateCamera'' function.
        solver: Flag from e.g openCV.SOLVEPNP_UPNP.
        distortion: Numpy array of shape of 5 elements calculated from
            the openCV calibrateCamera function.

    # Returns
        A list containing success flag, rotation and translation components
        of the 6D pose.
    """
    return cv2.solvePnP(points3D, points2D, camera.intrinsics,
                        camera.distortion, None, None, False, solver)


def project_points3D(points3D, pose6D, camera):
    """Projects 3D points into a specific pose.

    # Arguments
        points3D: Numpy array of shape ``(num_points, 3)``.
        pose6D: An instance of ``paz.abstract.Pose6D``.
        camera: An instance of ``paz.backend.Camera`` object.

    # Returns
        Numpy array of shape ``(num_points, 2)``
    """
    points2D, jacobian = cv2.projectPoints(
        points3D, pose6D.rotation_vector, pose6D.translation,
        camera.intrinsics, camera.distortion)
    # openCV adds a dimension to projection i.e. (num_points, 1, 2)
    points2D = np.squeeze(points2D, axis=1)
    return points2D


def project_to_image(rotation, translation, points3D, camera_intrinsics):
    """Project points3D to image plane using a perspective transformation.

              Image plane

           (0,0)-------->  (U)
             |
             |
             |
             v

            (V)

    # Arguments
        rotation: Array (3, 3). Rotation matrix (Rco).
        translation: Array (3). Translation (Tco).
        points3D: Array (num_points, 3). Points 3D in object frame.
        camera_intrinsics: Array of shape (3, 3). Diagonal elements represent
            focal lenghts and last column the image center translation.

    # Returns
        Array (num_points, 2) in UV image space.
    """
    if rotation.shape != (3, 3):
        raise ValueError('Rotation matrix is not of shape (3, 3)')
    if len(translation) != 3:
        raise ValueError('Translation vector is not of length 3')
    if len(points3D.shape) != 2:
        raise ValueError('Points3D should have a shape (num_points, 3)')
    if points3D.shape[1] != 3:
        raise ValueError('Points3D should have a shape (num_points, 3)')
    # TODO missing checks for camera intrinsics conditions
    points3D = np.matmul(rotation, points3D.T).T + translation
    x, y, z = np.split(points3D, 3, axis=1)
    x_focal_length = camera_intrinsics[0, 0]
    y_focal_length = camera_intrinsics[1, 1]
    x_image_center = camera_intrinsics[0, 2]
    y_image_center = camera_intrinsics[1, 2]
    x_points = (x_focal_length * (x / z)) + x_image_center
    y_points = (y_focal_length * (y / z)) + y_image_center
    projected_points2D = np.concatenate([x_points, y_points], axis=1)
    return projected_points2D


def translate_points2D_origin(points2D, coordinates):
    """Translates points2D to a different origin

    # Arguments
        points2D: Array (num_points, 2)
        coordinates: Array (4) containing (x_min, y_min, x_max, y_max)

    # Returns
        Translated points2D array (num_points, 2)
    """
    x_min, y_min, x_max, y_max = coordinates
    points2D[:, 0] = points2D[:, 0] + x_min
    points2D[:, 1] = points2D[:, 1] + y_min
    return points2D


def translate_keypoints(keypoints, translation):
    """Translate keypoints.

    # Arguments
        kepoints: Numpy array of shape ``(num_keypoints, 2)``.
        translation: A list of length two indicating the x,y translation values

    # Returns
        Numpy array
    """
    return keypoints + translation


def _preprocess_image_points2D(image_points2D):
    """Preprocessing image points for openCV's PnPRANSAC

    # Arguments
        image_points2D: Array of shape (num_points, 2)

    # Returns
        Contiguous float64 array of shape (num_points, 1, 2)
    """
    num_points = len(image_points2D)
    image_points2D = image_points2D.reshape(num_points, 1, 2)
    image_points2D = image_points2D.astype(np.float64)
    image_points2D = np.ascontiguousarray(image_points2D)
    return image_points2D


def solve_PnP_RANSAC(object_points3D, image_points2D, camera_intrinsics,
                     inlier_threshold=5, num_iterations=100):
    """Returns rotation (Roc) and translation (Toc) vectors that transform
        3D points in object frame to camera frame.

                               O------------O
                              /|           /|
                             / |          / |
                            O------------O  |
                            |  |    z    |  |
                            |  O____|____|__O
                            |  /    |___y|  /   object
                            | /    /     | /  coordinates
                            |/    x      |/
                            O------------O
                                   ___
                   Z                |
                  /                 | Rco, Tco
                 /_____X     <------|
                 |
                 |    camera
                 Y  coordinates

    # Arguments
        object_points3D: Array (num_points, 3). Points 3D in object reference
            frame. Represented as (0) in image above.
        image_points2D: Array (num_points, 2). Points in 2D in camera UV space.
        camera_intrinsics: Array of shape (3, 3). Diagonal elements represent
            focal lenghts and last column the image center translation.
        inlier_threshold: Number of inliers for RANSAC method.
        num_iterations: Maximum number of iterations.

    # Returns
        Rotation vector in axis-angle form (3) and translation vector (3).
    """
    if ((len(object_points3D) < 4) or (len(image_points2D) < 4)):
        raise ValueError('Solve PnP requires at least 4 3D and 2D points')
    image_points2D = _preprocess_image_points2D(image_points2D)
    success, rotation_vector, translation, inliers = cv2.solvePnPRansac(
        object_points3D, image_points2D, camera_intrinsics, None,
        flags=cv2.SOLVEPNP_EPNP, reprojectionError=inlier_threshold,
        iterationsCount=num_iterations)
    translation = np.squeeze(translation, 1)
    return success, rotation_vector, translation


def arguments_to_image_points2D(row_args, col_args):
    """Convert array arguments into UV coordinates.

              Image plane

           (0,0)-------->  (U)
             |
             |
             |
             v

            (V)

    # Arguments
        row_args: Array (num_rows).
        col_args: Array (num_cols).

    # Returns
        Array (num_cols, num_rows) representing points2D in UV space.

    # Notes
        Arguments are row args (V) and col args (U). Image points are in UV
            coordinates; thus, we concatenate them in that order
            i.e. [col_args, row_args]
    """
    row_args = row_args.reshape(-1, 1)
    col_args = col_args.reshape(-1, 1)
    image_points2D = np.concatenate([col_args, row_args], axis=1)  # (U, V)
    return image_points2D


def normalize_keypoints(keypoints, height, width):
    """Transform keypoints in image coordinates to normalized coordinates
    # Arguments
        keypoints: Numpy array of shape ``(num_keypoints, 2)``.
        height: Int. Height of the image
        width: Int. Width of the image
    # Returns
        Numpy array of shape ``(num_keypoints, 2)``.
    """
    warn('DEPRECATED please use denomarlize_points2D')
    normalized_keypoints = np.zeros_like(keypoints, dtype=np.float32)
    for keypoint_arg, keypoint in enumerate(keypoints):
        x, y = keypoint[:2]
        # transform key-point coordinates to image coordinates
        x = (((x + 0.5) - (width / 2.0)) / (width / 2))
        y = (((height - 0.5 - y) - (height / 2.0)) / (height / 2))
        normalized_keypoints[keypoint_arg][:2] = [x, y]
    return normalized_keypoints


def denormalize_keypoints(keypoints, height, width):
    """Transform normalized keypoint coordinates into image coordinates
    # Arguments
        keypoints: Numpy array of shape ``(num_keypoints, 2)``.
        height: Int. Height of the image
        width: Int. Width of the image
    # Returns
        Numpy array of shape ``(num_keypoints, 2)``.
    """
    warn('DEPRECATED please use denomarlize_points2D')
    for keypoint_arg, keypoint in enumerate(keypoints):
        x, y = keypoint[:2]
        # transform key-point coordinates to image coordinates
        x = (min(max(x, -1), 1) * width / 2 + width / 2) - 0.5
        # flip since the image coordinates for y are flipped
        y = height - 0.5 - (min(max(y, -1), 1) * height / 2 + height / 2)
        x, y = int(round(x)), int(round(y))
        keypoints[keypoint_arg][:2] = [x, y]
    return keypoints


def rotate_point2D(point2D, rotation_angle):
    """Rotate keypoint.

    # Arguments
        point2D: keypoint [x, y]
        rotation angle: Int. Angle of rotation.

    # Returns
        List of x and y rotated points
    """
    rotation_angle = np.pi * rotation_angle / 180
    sin_n, cos_n = np.sin(rotation_angle), np.cos(rotation_angle)
    x_rotated = (point2D[0] * cos_n) - (point2D[1] * sin_n)
    y_rotated = (point2D[0] * sin_n) + (point2D[1] * cos_n)
    return [x_rotated, y_rotated]


def transform_keypoint(keypoint, transform):
    """ Transform keypoint.

    # Arguments
        keypoint2D: keypoint [x, y]
        transform: Array. Transformation matrix
    """
    keypoint = np.array([keypoint[0], keypoint[1], 1.]).T
    transformed_keypoint = np.dot(transform, keypoint)
    return transformed_keypoint


def add_offset_to_point(keypoint_location, offset=0):
    """ Add offset to keypoint location

    # Arguments
        keypoint_location: keypoint [y, x]
        offset: Float.
    """
    y, x = keypoint_location
    y = y + offset
    x = x + offset
    return y, x


def flip_keypoints_left_right(keypoints, image_size=(32, 32)):
    """Flip the detected 2D keypoints left to right.

    # Arguments
        keypoints: Array
        image_size: list/tuple
        axis: int

    # Returns
        flipped_keypoints: Numpy array
    """
    x_coordinates, y_coordinates = np.split(keypoints, 2, axis=1)
    flipped_x = image_size[0] - x_coordinates
    keypoints = np.concatenate((flipped_x, y_coordinates), axis=1)
    return keypoints


def compute_orientation_vector(keypoints3D, parents):
    """Compute bone orientations from joint coordinates
       (child joint - parent joint). The returned vectors are normalized.
       For the root joint, it will be a zero vector.

    # Arguments
        keypoints3D : Numpy array [num_keypoints, 3]. Joint coordinates.
        parents: Parents of the keypoints from kinematic chain

    # Returns
        Array [num_keypoints, 3]. The unit vectors from each child joint to
        its parent joint. For the root joint, it's are zero vector.
    """
    delta = []
    for joint_arg in range(len(parents)):
        parent = parents[joint_arg]
        if parent is None:
            delta.append(np.zeros(3))
        else:
            delta.append(keypoints3D[joint_arg] - keypoints3D[parent])
    delta = np.stack(delta, 0)
    return delta


def rotate_keypoints3D(rotation_matrix, keypoints):
    """Rotatate the keypoints by using rotation matrix

    # Arguments
        Rotation matrix [N, 3, 3].
        keypoints [N, 3]

    # Returns
        Rotated keypoints [N, 3]
    """
    keypoint_xyz = np.einsum('ijk, ik -> ij', rotation_matrix, keypoints)
    return keypoint_xyz


def flip_along_x_axis(keypoints, axis=0):
    """Flip the keypoints along the x axis.

    # Arguments
        keypoints: Array
        axis: int/list

    # Returns
        Flipped keypoints: Array
    # """
    x, y, z = np.split(keypoints, 3, axis=1)
    keypoints = np.concatenate((-x, y, z), axis=1)
    return keypoints


def uv_to_vu(keypoints):
    """Flips the uv coordinates to vu.

    # Arguments
        keypoints: Array.
    """
    return keypoints[:, ::-1]


def standardize(data, mean, scale):
    """it takes the data the mean and the standard deviation
       and returns the standardized data

    # Arguments
        data: nxd matrix to normalize
        mean: Array of means
        scale: standard deviation

    # Returns
        standardized poses2D
    # """
    return np.divide((data - mean), scale)


def destandardize(data, mean, scale):
    """it takes the standardized data the mean and the standard
       deviation and returns the destandardized data

    # Arguments
        data: nxd matrix to unnormalize
        mean: Array of means
        scale: standard deviation

    # Returns
        destandardized poses3D
    """
    return (data * scale) + mean


def initialize_translation(joints2D, camera_intrinsics, ratio):
    """Computes initial 3D translation of root joint

    # Arguments
        joints2D: 2D root joint from HigherHRNet
        camera_intrinsics: camera intrinsic parameters
        ratio: ration of sum of 3D bones to 2D bones

    # Returns
        Array of initial estimate of the global position
        of the root joint in 3D
    """
    focal_length = camera_intrinsics[0, 0]
    image_center_x = camera_intrinsics[0, 2]
    image_center_y = camera_intrinsics[1, 2]
    z = focal_length * ratio
    x = (joints2D[:, 0] - image_center_x) * ratio
    y = (joints2D[:, 1] - image_center_y) * ratio
    translation = np.array((x, y, z))
    return translation.flatten()


def solve_least_squares(solver, compute_joints_distance,
                        initial_joints_translation, joints3D,
                        poses2D, camera_intrinsics):
    """ Solve the least squares

    # Arguments
        solver: from scipy.optimize import least_squares
        compute_joints_distance: global_pose.compute_joints_distance
        initial_root_translation: initial 3D translation of root joint
        joints3D: 16 moving joints in 3D
        poses2d: 2D poses
        camera_intrinsics: camera intrinsic parameters

    Returns
        optimal translation of root joint for each person
    """
    joints_translation = solver(
        compute_joints_distance, initial_joints_translation, verbose=0,
        args=(joints3D, poses2D, camera_intrinsics))
    joints_translation = np.reshape(joints_translation.x, (-1, 3))
    return joints_translation


def get_bones_length(poses2D, poses3D):
    """Computes sum of bone lengths in 3D

    #Arguments
        poses3D: array of predicted poses in 3D (Nx16x3)
        poses2D: array of poses in 2D    (Nx32)

    #Returns
        sum_bones2D: sum of length of all bones in the 3D skeleton
        sum_bones3D: sum of length of all bones in the 3D skeleton
    """
    sum_bones2D = 0
    sum_bones3D = np.zeros(poses3D.shape[0])
    start_joints = np.arange(0, 15)
    end_joints = np.arange(1, 16)
    for person in poses2D:
        bone_length = np.linalg.norm(person[start_joints] -
                                     person[end_joints])
        sum_bones2D = sum_bones2D + bone_length
    for person in poses3D:
        bone_length = np.linalg.norm(person[start_joints] -
                                     person[end_joints])
        sum_bones3D = sum_bones3D + bone_length
    return sum_bones2D, sum_bones3D


def compute_reprojection_error(initial_translation, keypoints3D,
                               keypoints2D, camera_intrinsics):
    """compute distance between each person joints

    # Arguments
        initial_translation: initial guess of position of joint
        keypoints3D: 3D keypoints to be optimized (Nx16x3)
        keypoints2D: 2D keypoints (Nx32)
        camera_inrinsics: camera intrinsic parameters

    # Returns
        person_sum: sum of L2 distances between each joint per person
    """
    initial_translation = np.reshape(initial_translation, (-1, 3))
    new_poses3D = np.zeros((keypoints3D.shape))
    for person in range(len(initial_translation)):
        new_poses3D[person] = (keypoints3D[person] +
                               initial_translation[person])
    new_poses3D = new_poses3D.reshape((-1, 3))
    rotation = np.identity(3)
    translation = np.zeros((3,))
    project2D = project_to_image(rotation, translation, new_poses3D,
                                 camera_intrinsics)
    joints_distance = np.linalg.norm(np.ravel(keypoints2D) -
                                     np.ravel(project2D))

    return np.sum(joints_distance)


def merge_into_mean(keypoints2D, args_to_mean):
    """merge keypoints and take the mean

    # Arguments:
             keypoints2D: keypoints2D (Nx17x2)
             args_to_mean: dict of joint indices

    # Returns:
             keypoints2D: keypoints2D after merging
            """
    for point, joints_indices in args_to_mean.items():
        keypoints2D[:, point] = (keypoints2D[:, joints_indices[0]] +
                                 keypoints2D[:, joints_indices[1]]) / 2
    return keypoints2D


def filter_keypoints(keypoints, args_to_joints):
    """filter keypoints.

    # Arguments
        keypoints: points in camera coordinates
        args_to_joints: Array of joints indices

    # Returns
        filtered keypoints
    # """
    return keypoints[:, args_to_joints, :]


def filter_keypoints3D(keypoints3D, args_to_joints3D):
    """Selects 16 moving joints (Neck/Nose excluded) from 32 predicted
       joints in 3D

    # Arguments
        keypoints3D: Nx96 points in camera coordinates
        args_to_joints3D: list of indices

    # Returns
        filtered_joints_3D: Nx48 points (moving joints)
    """
    keypoints_num = len(keypoints3D)
    keypoints3D = np.reshape(keypoints3D, [keypoints_num, 32, 3])
    joints3D = filter_keypoints(keypoints3D, args_to_joints3D)
    return joints3D


def filter_keypoints2D(keypoints2D, args_to_mean, h36m_to_coco_joints2D):
    """Selects 16 moving joints (Neck/Nose excluded) from 17 predicted
       joints in 2D

    # Arguments
        keypoints3D: Nx17x2 points in camera coordinates
        args_to_mean: keypoints indices
        h36m_to_coco_joints2D: human36m dataset list of joints indices

    # Returns
        joints2D: Nx32 points (moving joints)
    """
    joints2D = filter_keypoints(keypoints2D, h36m_to_coco_joints2D)
    joints2D = np.reshape(joints2D, [joints2D.shape[0], -1])
    return joints2D


def compute_optimized_pose3D(keypoints3D, joint_translation,
                             camera_intrinsics):
    """Compute the optimized 3D pose

    # Arguments
        keypoints3D: 3D keypoints
        joint_translation: np array joints translation
        camera_intrinsics: camera intrinsics parameters

    # Returns
        optimized_poses3D: np array of optimized posed3D
    """
    optimized_poses3D = []
    for person in range(keypoints3D.shape[0]):
        keypoints3D[person] = keypoints3D[person] + joint_translation[person]
        rotation = np.identity(3)
        translation = np.zeros((3,))
        points = project_to_image(rotation, translation,
                                  keypoints3D[person].reshape((-1, 3)),
                                  camera_intrinsics)
        optimized_poses3D.append(np.reshape(points, [1, 64]))
    return np.array(optimized_poses3D)



import math
import tensorflow as tf
import numpy as np

from ...models.keypoint.projector import Projector


class KeypointNetLoss(object):
    """KeypointNet loss for discovering latent keypoints.

    # Arguments
        num_keypints: Int. Number of keypoints to discover.
        focal_length: Float. Focal length of camera
        rotation_noise: Float. Noise added to the estimation of the rotation.
        separation_delta: Float. Delta used for the ''separation'' loss.
        loss_weights: Dict. having as keys strings with the different losses
            names e.g. ''consistency'' and as value the weight used for that
            loss.

    # References
        - [Discovery of Latent 3D Keypoints via End-to-end
            Geometric Reasoning](https://arxiv.org/pdf/1807.03146.pdf)
    """
    def __init__(self, num_keypoints, focal_length, rotation_noise=0.1,
                 separation_delta=0.05, loss_weights={
                     'consistency': 1.0, 'silhouette': 1.0, 'separation': 1.0,
                     'relative_pose': 0.2, 'variance': 0.5}):

        self.num_keypoints = int(num_keypoints)
        self.focal_length = focal_length
        self.projector = Projector(focal_length)
        self.rotation_noise = rotation_noise
        self.separation_delta = separation_delta
        self.loss_weights = loss_weights

    def _reshape_matrix(self, matrix):
        matrix = tf.reshape(matrix, [-1, 4, 4])
        # transpose is for multiplying points with matrices from the left.
        matrix = tf.transpose(matrix, [0, 2, 1])
        return matrix

    def _unpack_matrices(self, matrices):
        world_to_A = self._reshape_matrix(matrices[:, 0, :])
        world_to_B = self._reshape_matrix(matrices[:, 1, :])
        A_to_world = self._reshape_matrix(matrices[:, 2, :])
        B_to_world = self._reshape_matrix(matrices[:, 3, :])
        return world_to_A, world_to_B, A_to_world, B_to_world

    def _unpack_uvz_coordinates(self, uvz_coordinates):
        uvz_A = uvz_coordinates[:, :self.num_keypoints, :]
        uvz_B = uvz_coordinates[:, self.num_keypoints:, :]
        return uvz_A, uvz_B

    def _consistency(self, uvz_M, M_to_world, world_to_N, uvz_N):
        keypoints_M = self.projector.unproject(uvz_M)
        world_coordinates = tf.matmul(keypoints_M, M_to_world)
        keypoints_M_in_N = tf.matmul(world_coordinates, world_to_N)
        uvz_M_in_N = self.projector.project(keypoints_M_in_N)
        squared_difference = tf.square(uvz_M_in_N - uvz_N)
        l2_distance = tf.reduce_sum(squared_difference, axis=[1, 2])
        consistency_loss = l2_distance / self.num_keypoints
        return consistency_loss

    def consistency(self, matrices, uvz_coordinates):
        matrices = self._unpack_matrices(matrices)
        world_to_A, world_to_B, A_to_world, B_to_world = matrices
        uvz_A, uvz_B = self._unpack_uvz_coordinates(uvz_coordinates)
        consistency_A = self._consistency(uvz_A, A_to_world, world_to_B, uvz_B)
        consistency_B = self._consistency(uvz_B, B_to_world, world_to_A, uvz_A)
        consistency_loss = (consistency_A + consistency_B) / 2.0
        consistency_loss = self.loss_weights['consistency'] * consistency_loss
        return consistency_loss

    def _separation(self, uvz):
        keypoints = self.projector.unproject(uvz)
        keypoints_i = tf.tile(keypoints, [1, self.num_keypoints, 1])
        keypoints_j = tf.tile(keypoints, [1, 1, self.num_keypoints])
        keypoints_j = tf.reshape(keypoints_j, tf.shape(keypoints_i))
        squared_difference = tf.square(keypoints_i - keypoints_j)
        squared_l2_distance = tf.reduce_sum(squared_difference, axis=2)
        separation_loss = tf.maximum(
            -squared_l2_distance + self.separation_delta, 0.0)
        separation_loss = tf.reshape(
            separation_loss, [-1, self.num_keypoints, self.num_keypoints])
        separation_loss = tf.reduce_sum(separation_loss, axis=[1, 2])
        separation_loss = separation_loss / self.num_keypoints
        return separation_loss

    def separation(self, matrices, uvz_coordinates):
        uvz_A, uvz_B = self._unpack_uvz_coordinates(uvz_coordinates)
        separation_loss_A = self._separation(uvz_A)
        separation_loss_B = self._separation(uvz_B)
        separation_loss = (separation_loss_A + separation_loss_B) / 2.0
        separation_loss = self.loss_weights['separation'] * separation_loss
        return separation_loss

    def relative_pose(self, matrices, uvz_coordinates):
        matrices = self._unpack_matrices(matrices)
        world_to_A, world_to_B, A_to_world, B_to_world = matrices
        uvz_A, uvz_B = self._unpack_uvz_coordinates(uvz_coordinates)
        keypoints_A = self.projector.unproject(uvz_A)
        keypoints_B = self.projector.unproject(uvz_B)

        A_to_B = tf.matmul(A_to_world, world_to_B)
        rotation_A_to_B = A_to_B[:, :3, :3]
        estimation_args = (keypoints_A, keypoints_B, self.rotation_noise)
        estimated_rotation_A_to_B = self.estimate_rotation(*estimation_args)
        estimated_rotation_A_to_B = estimated_rotation_A_to_B[:, :3, :3]
        squared_A_to_B = tf.square(estimated_rotation_A_to_B - rotation_A_to_B)
        squared_frobenius = tf.reduce_sum(squared_A_to_B, axis=[1, 2])
        frobenius = tf.sqrt(squared_frobenius)
        arcsin_arg = tf.minimum(1.0, frobenius / (2 * math.sqrt(2)))
        angular_loss = 2.0 * tf.asin(arcsin_arg)
        angular_loss = self.loss_weights['relative_pose'] * angular_loss
        return angular_loss

    def uvz_points(self, matrices, uvz_coordinates):
        consistency_loss = self.consistency(matrices, uvz_coordinates)
        separation_loss = self.separation(matrices, uvz_coordinates)
        relative_pose_loss = self.relative_pose(matrices, uvz_coordinates)
        uvz_loss = consistency_loss + separation_loss + relative_pose_loss
        return uvz_loss

    def _silhouette(self, alpha_channel, uv_volume):
        alpha_channel = tf.greater(alpha_channel, tf.zeros_like(alpha_channel))
        alpha_channel = tf.cast(alpha_channel, dtype=tf.float32)
        alpha_channel = tf.expand_dims(alpha_channel, 1)
        silhouette_loss = tf.reduce_sum(uv_volume * alpha_channel, axis=[2, 3])
        silhouette_loss = -tf.math.log(silhouette_loss + 1e-12)
        silhouette_loss = tf.reduce_mean(silhouette_loss, axis=-1)
        return silhouette_loss

    def silhouette(self, alpha_channels, uv_volumes):
        alpha_channel_A = alpha_channels[:, :, :, 0]
        alpha_channel_B = alpha_channels[:, :, :, 1]
        uv_volume_A = uv_volumes[:, :self.num_keypoints, :, :]
        uv_volume_B = uv_volumes[:, self.num_keypoints:, :, :]
        silhouette_loss_A = self._silhouette(alpha_channel_A, uv_volume_A)
        silhouette_loss_B = self._silhouette(alpha_channel_B, uv_volume_B)
        silhouette_loss = (silhouette_loss_A + silhouette_loss_B) / 2.0
        silhouette_loss = self.loss_weights['silhouette'] * silhouette_loss
        return silhouette_loss

    def _variance(self, uv_volume, range_u, range_v):
        expected_keypoint_u = tf.reduce_sum(uv_volume * range_u, axis=[2, 3])
        expected_keypoint_v = tf.reduce_sum(uv_volume * range_v, axis=[2, 3])
        uv = tf.stack([expected_keypoint_u, expected_keypoint_v], -1)
        uv = tf.reshape(uv, [-1, self.num_keypoints, 2])
        uv = tf.reshape(uv, [tf.shape(uv)[0], tf.shape(uv)[1], 1, 1, 2])

        ranges = tf.stack([range_u, range_v], axis=2)
        ranges_sh = tf.shape(ranges)
        ranges = tf.reshape(ranges, [1, 1, ranges_sh[0], ranges_sh[1], 2])
        squared_difference = tf.reduce_sum(tf.square(uv - ranges), axis=4)
        diff = squared_difference * uv_volume
        diff = tf.reduce_sum(diff, axis=[2, 3])
        variance = tf.reduce_mean(diff, axis=-1)
        return variance

    def variance(self, alpha_channels, uv_volumes):
        uv_volume_A = uv_volumes[:, :self.num_keypoints, :, :]
        uv_volume_B = uv_volumes[:, self.num_keypoints:, :, :]
        feature_map_size = uv_volumes.shape[-1]

        arange = np.arange(0.5, feature_map_size, 1)
        arange = arange / (feature_map_size / 2) - 1
        range_u, range_v = tf.meshgrid(arange, -arange)
        range_u = tf.cast(range_u, dtype=tf.float32)
        range_v = tf.cast(range_v, dtype=tf.float32)

        variance_loss_A = self._variance(uv_volume_A, range_u, range_v)
        variance_loss_B = self._variance(uv_volume_B, range_u, range_v)
        variance_loss = (variance_loss_A + variance_loss_B) / 2.0
        variance_loss = self.loss_weights['variance'] * variance_loss
        return variance_loss

    def uv_volumes(self, alpha_channels, uv_volumes):
        variance_loss = self.variance(alpha_channels, uv_volumes)
        silhouette_loss = self.silhouette(alpha_channels, uv_volumes)
        uv_loss = variance_loss + silhouette_loss
        return uv_loss

    def estimate_rotation(self, keypoints_A, keypoints_B, noise=0.1):
        """Estimates the rotation between two sets of keypoints using
        Kabsch algorithm.

        The rotation is estimated by first subtracting mean from each
        set of keypoints and computing SVD of the covariance matrix.

        Arguments:
            xyz0: [batch, num_kp, 3] The first set of keypoints.
            xyz1: [batch, num_kp, 3] The second set of keypoints.
            pconf: [batch, num_kp] The weights used to
                   compute the rotation estimate.
            noise: A number indicating the noise added to the keypoints.

        Returns:
            [batch, 3, 3] A batch of transposed 3 x 3 rotation matrices.
        """

        pconf = tf.ones(
            [tf.shape(keypoints_A)[0],
             tf.shape(keypoints_A)[1]],
            dtype=tf.float32) / self.num_keypoints

        noise_A = tf.random.normal(tf.shape(keypoints_A), mean=0, stddev=noise)
        noise_B = tf.random.normal(tf.shape(keypoints_B), mean=0, stddev=noise)
        keypoints_A = keypoints_A + noise_A
        keypoints_B = keypoints_B + noise_B
        pconf2 = tf.expand_dims(pconf, 2)
        center_A = tf.reduce_sum(pconf2 * keypoints_A, 1, keepdims=True)
        center_B = tf.reduce_sum(pconf2 * keypoints_B, 1, keepdims=True)
        x = keypoints_A - center_A
        y = keypoints_B - center_B
        weighted_x = tf.matmul(
            x, tf.linalg.diag(pconf), transpose_a=True)
        covariance = tf.matmul(weighted_x, y)
        _, u, v = tf.linalg.svd(covariance, full_matrices=True)
        d = tf.linalg.det(tf.matmul(v, u, transpose_b=True))
        ud = tf.concat(
            [u[:, :, :-1],
             u[:, :, -1:] * tf.expand_dims(tf.expand_dims(d, 1), 1)],
            axis=2)
        return tf.matmul(ud, v, transpose_b=True)



import numpy as np

from ..abstract import Processor

from ..backend.image import cast_image
from ..backend.image import load_image
from ..backend.image import random_saturation
from ..backend.image import random_brightness
from ..backend.image import random_contrast
from ..backend.image import random_hue
from ..backend.image import resize_image
from ..backend.image import scale_resize
from ..backend.image import random_image_blur
from ..backend.image import random_flip_left_right
from ..backend.image import convert_color_space
from ..backend.image import show_image
from ..backend.image import blend_alpha_channel
from ..backend.image import random_shape_crop
from ..backend.image import make_random_plain_image
from ..backend.image import concatenate_alpha_mask
from ..backend.image import draw_filled_polygon
from ..backend.image import gaussian_image_blur
from ..backend.image import normalized_device_coordinates_to_image
from ..backend.image import image_to_normalized_device_coordinates
from ..backend.image import replace_lower_than_threshold
from ..backend.image import flip_left_right
from ..backend.image import BILINEAR
from ..backend.image.tensorflow_image import imagenet_preprocess_input


B_IMAGENET_MEAN, G_IMAGENET_MEAN, R_IMAGENET_MEAN = 104, 117, 123
BGR_IMAGENET_MEAN = (B_IMAGENET_MEAN, G_IMAGENET_MEAN, R_IMAGENET_MEAN)
RGB_IMAGENET_MEAN = (R_IMAGENET_MEAN, G_IMAGENET_MEAN, B_IMAGENET_MEAN)
B_IMAGENET_STDEV, G_IMAGENET_STDEV, R_IMAGENET_STDEV = 57.3, 57.1, 58.4
RGB_IMAGENET_STDEV = (R_IMAGENET_STDEV, G_IMAGENET_STDEV, B_IMAGENET_STDEV)


class CastImage(Processor):
    """Cast image to given dtype.

    # Arguments
        dtype: Str or np.dtype
    """
    def __init__(self, dtype):
        self.dtype = dtype
        super(CastImage, self).__init__()

    def call(self, image):
        return cast_image(image, self.dtype)


class SubtractMeanImage(Processor):
    """Subtract channel-wise mean to image.

    # Arguments
        mean: List of length 3, containing the channel-wise mean.
    """
    def __init__(self, mean):
        self.mean = mean
        super(SubtractMeanImage, self).__init__()

    def call(self, image):
        return image - self.mean


class AddMeanImage(Processor):
    """Adds channel-wise mean to image.

    # Arguments
        mean: List of length 3, containing the channel-wise mean.
    """
    def __init__(self, mean):
        self.mean = mean
        super(AddMeanImage, self).__init__()

    def call(self, image):
        return image + self.mean


class NormalizeImage(Processor):
    """Normalize image by diving all values by 255.0.
    """
    def __init__(self):
        super(NormalizeImage, self).__init__()

    def call(self, image):
        return image / 255.0


class DenormalizeImage(Processor):
    """Denormalize image by multiplying all values by 255.0.
    """
    def __init__(self):
        super(DenormalizeImage, self).__init__()

    def call(self, image):
        return image * 255.0


class LoadImage(Processor):
    """Loads image.

    # Arguments
        num_channels: Integer, valid integers are: 1, 3 and 4.
    """
    def __init__(self, num_channels=3):
        self.num_channels = num_channels
        super(LoadImage, self).__init__()

    def call(self, image):
        return load_image(image, self.num_channels)


class RandomSaturation(Processor):
    """Applies random saturation to an image in RGB space.

    # Arguments
        lower: Float, lower bound for saturation factor.
        upper: Float, upper bound for saturation factor.
    """
    def __init__(self, lower=0.3, upper=1.5):
        self.lower = lower
        self.upper = upper
        super(RandomSaturation, self).__init__()

    def call(self, image):
        return random_saturation(image, self.lower, self.upper)


class RandomBrightness(Processor):
    """Adjust random brightness to an image in RGB space.

    # Arguments
        max_delta: Float.
    """
    def __init__(self, delta=32):
        self.delta = delta
        super(RandomBrightness, self).__init__()

    def call(self, image):
        return random_brightness(image, self.delta)


class RandomContrast(Processor):
    """Applies random contrast to an image in RGB

    # Arguments
        lower: Float, indicating the lower bound of the random number
            to be multiplied with the BGR/RGB image.
        upper: Float, indicating the upper bound of the random number
        to be multiplied with the BGR/RGB image.
    """
    def __init__(self, lower=0.5, upper=1.5):
        self.lower = lower
        self.upper = upper
        super(RandomContrast, self).__init__()

    def call(self, image):
        return random_contrast(image, self.lower, self.upper)


class RandomHue(Processor):
    """Applies random hue to an image in RGB space.

    # Arguments
        delta: Int, indicating the range (-delta, delta ) of possible
            hue values.
    """
    def __init__(self, delta=18):
        self.delta = delta
        super(RandomHue, self).__init__()

    def call(self, image):
        return random_hue(image, self.delta)


class ResizeImage(Processor):
    """Resize image.

    # Arguments
        size: List of two ints.
    """
    def __init__(self, shape, method=BILINEAR):
        self.shape = shape
        self.method = method
        super(ResizeImage, self).__init__()

    def call(self, image):
        return resize_image(image, self.shape, self.method)


class ResizeImages(Processor):
    """Resize list of images.

    # Arguments
        size: List of two ints.
    """
    def __init__(self, shape):
        self.shape = shape
        super(ResizeImages, self).__init__()

    def call(self, images):
        return [resize_image(image, self.shape) for image in images]


class RandomImageBlur(Processor):
    """Randomizes image quality

    # Arguments
        probability: Float between [0, 1]. Assigns probability of how
            often a random image blur is applied.
    """
    def __init__(self, probability=0.5):
        super(RandomImageBlur, self).__init__()
        self.probability = probability

    def call(self, image):
        if self.probability >= np.random.rand():
            image = random_image_blur(image)
        return image


class RandomGaussianBlur(Processor):
    """Randomizes image quality

    # Arguments
        probability: Float between [0, 1]. Assigns probability of how
            often a random image blur is applied.
    """
    def __init__(self, kernel_size=(5, 5), probability=0.5):
        super(RandomGaussianBlur, self).__init__()
        self.kernel_size = kernel_size
        self.probability = probability

    def call(self, image):
        if self.probability >= np.random.rand():
            image = gaussian_image_blur(image, self.kernel_size)
        return image


class RandomFlipImageLeftRight(Processor):
    """Randomly flip the image left or right
    """
    def __init__(self):
        super(RandomFlipImageLeftRight, self).__init__()

    def call(self, image):
        return random_flip_left_right(image)


class ConvertColorSpace(Processor):
    """Converts image to a different color space.

    # Arguments
        flag: Flag found in ``processors``indicating transform e.g.
            ``pr.BGR2RGB``
    """
    def __init__(self, flag):
        self.flag = flag
        super(ConvertColorSpace, self).__init__()

    def call(self, image):
        return convert_color_space(image, self.flag)


class ShowImage(Processor):
    """Shows image in a separate window.

    # Arguments
        window_name: String. Window name.
        wait: Boolean
    """
    def __init__(self, window_name='image', wait=True):
        super(ShowImage, self).__init__()
        self.window_name = window_name
        self.wait = wait

    def call(self, image):
        return show_image(image, self.window_name, self.wait)


class ImageDataProcessor(Processor):
    """Wrapper for Keras ImageDataGenerator

    # Arguments
        generator: An instantiated Keras ImageDataGenerator
    """
    def __init__(self, generator):
        super(ImageDataProcessor, self).__init__()
        self.generator = generator

    def call(self, image):
        random_parameters = self.generator.get_random_transform(image.shape)
        image = self.generator.apply_transform(image, random_parameters)
        image = self.generator.standardize(image)
        return image


class AlphaBlending(Processor):
    """Blends image to background using the image's alpha channel.
    """
    def __init__(self):
        super(AlphaBlending, self).__init__()

    def call(self, image, background):
        return blend_alpha_channel(image, background)


class RandomShapeCrop(Processor):
    """Randomly crops a part of an image of always the same given ``shape``.

    # Arguments
        shape: List of two ints [height, width].
            Dimensions of image to be cropped.
    """
    def __init__(self, shape):
        super(RandomShapeCrop, self).__init__()
        self.shape = shape

    def call(self, image):
        return random_shape_crop(image, self.shape)


class MakeRandomPlainImage(Processor):
    """Makes random plain image by randomly sampling an RGB color.

    # Arguments
        shape: List of two ints [height, width].
            Dimensions of plain image to be generated.
    """
    def __init__(self, shape):
        super(MakeRandomPlainImage, self).__init__()
        self.shape = shape

    def call(self):
        return make_random_plain_image(self.shape)


class ConcatenateAlphaMask(Processor):
    """Concatenates alpha mask to original image.
    """
    def __init__(self, **kwargs):
        super(ConcatenateAlphaMask, self).__init__(**kwargs)

    def call(self, image, alpha_mask):
        return concatenate_alpha_mask(image, alpha_mask)


class BlendRandomCroppedBackground(Processor):
    """Blends image with a randomly cropped background.

    # Arguments
        background_paths: List of strings. Each element of the list is a
            full-path to an image used for cropping a background.
    """
    def __init__(self, background_paths):
        super(BlendRandomCroppedBackground, self).__init__()
        if not isinstance(background_paths, list):
            raise ValueError('``background_paths`` must be list')
        if len(background_paths) == 0:
            raise ValueError('No paths given in ``background_paths``')
        self.background_paths = background_paths

    def call(self, image):
        random_arg = np.random.randint(0, len(self.background_paths))
        background_path = self.background_paths[random_arg]
        background = load_image(background_path)
        background = random_shape_crop(background, image.shape[:2])
        if background is None:
            H, W, num_channels = image.shape
            # background contains always a channel less
            num_channels = num_channels - 1
            background = make_random_plain_image((H, W, num_channels))
        return blend_alpha_channel(image, background)


class AddOcclusion(Processor):
    """Adds a random occlusion to image by generating random vertices and
        drawing a polygon.

    # Arguments
        max_radius_scale: Float between [0, 1].
            Value multiplied with largest image dimension to obtain the maximum
                radius possible of a vertex in the occlusion polygon.
        probability: Float between [0, 1]. Assigns probability of how
            often an occlusion to an image is generated.
    """
    def __init__(self, max_radius_scale=0.5, probability=0.5):
        super(AddOcclusion, self).__init__()
        self.max_radius_scale = max_radius_scale
        self.probability = probability

    def _random_vertices(self, center, max_radius, min_vertices, max_vertices):
        num_vertices = np.random.randint(min_vertices, max_vertices)
        angle_delta = 2 * np.pi / num_vertices
        initial_angle = np.random.uniform(0, 2 * np.pi)
        angles = initial_angle + np.arange(0, num_vertices) * angle_delta
        x_component = np.cos(angles).reshape(-1, 1)
        y_component = np.sin(angles).reshape(-1, 1)
        vertices = np.concatenate([x_component, y_component], -1)
        random_lengths = np.random.uniform(0, max_radius, num_vertices)
        random_lengths = random_lengths.reshape(num_vertices, 1)
        vertices = vertices * random_lengths
        vertices = vertices + center
        return vertices.astype(np.int32)

    def add_occlusion(self, image, max_radius_scale):
        height, width = image.shape[:2]
        max_radius = np.max((height, width)) * max_radius_scale
        center = np.random.rand(2) * np.array([width, height])
        vertices = self._random_vertices(center, max_radius, 3, 7)
        color = np.random.randint(0, 256, 3).tolist()
        return draw_filled_polygon(image, vertices, color)

    def call(self, image):
        if self.probability >= np.random.rand():
            image = self.add_occlusion(image, self.max_radius_scale)
        return image


class RandomImageCrop(Processor):
    """Crops randomly a rectangle from an image.

    # Arguments
        crop_factor: Float between ``[0, 1]``.
        probability: Float between ``[0, 1]``.
    """
    def __init__(self, crop_factor=0.3, probability=0.5):
        self.crop_factor = crop_factor
        self.probability = probability
        super(RandomImageCrop, self).__init__()

    def call(self, image):
        if self.probability < np.random.rand():
            return image
        H, W = image.shape[:2]
        W_crop = np.random.uniform(self.crop_factor * W, W)
        H_crop = np.random.uniform(self.crop_factor * H, H)
        x_min = np.random.uniform(W - W_crop)
        y_min = np.random.uniform(H - H_crop)
        x_max = x_min + W_crop
        y_max = y_min + H_crop
        cropped_image = image[int(x_min):int(x_max), int(y_min):int(y_max), :]
        return cropped_image


class ImageToNormalizedDeviceCoordinates(Processor):
    """Map image value from [0, 255] -> [-1, 1].
    """
    def __init__(self):
        super(ImageToNormalizedDeviceCoordinates, self).__init__()

    def call(self, image):
        return image_to_normalized_device_coordinates(image)


class NormalizedDeviceCoordinatesToImage(Processor):
    """Map normalized value from [-1, 1] -> [0, 255].
    """
    def __init__(self):
        super(NormalizedDeviceCoordinatesToImage, self).__init__()

    def call(self, image):
        return normalized_device_coordinates_to_image(image)


class ReplaceLowerThanThreshold(Processor):
    def __init__(self, threshold=1e-8, replacement=0.0):
        super(ReplaceLowerThanThreshold, self).__init__()
        self.threshold = threshold
        self.replacement = replacement

    def call(self, values):
        return replace_lower_than_threshold(
            values, self.threshold, self.replacement)


class GetNonZeroValues(Processor):
    def __init__(self):
        super(GetNonZeroValues, self).__init__()

    def call(self, array):
        channel_wise_sum = np.sum(array, axis=2)
        non_zero_arguments = np.nonzero(channel_wise_sum)
        return array[non_zero_arguments]


class GetNonZeroArguments(Processor):
    def __init__(self):
        super(GetNonZeroArguments, self).__init__()

    def call(self, array):
        channel_wise_sum = np.sum(array, axis=2)
        non_zero_rows, non_zero_columns = np.nonzero(channel_wise_sum)
        return non_zero_rows, non_zero_columns


class ImagenetPreprocessInput(Processor):
    def __init__(self):
        super(ImagenetPreprocessInput, self).__init__()

    def call(self, image):
        return imagenet_preprocess_input(image)


class FlipLeftRightImage(Processor):
    """Flips an image left and right.

    # Arguments
        image: Numpy array.
    """
    def __init__(self):
        super(FlipLeftRightImage, self).__init__()

    def call(self, image):
        return flip_left_right(image)


class DivideStandardDeviationImage(Processor):
    """Divide channel-wise standard deviation to image.

    # Arguments
        standard_deviation: List of length 3, containing the
            channel-wise standard deviation.

    # Properties
        standard_deviation: List.

    # Methods
        call()
    """
    def __init__(self, standard_deviation):
        self.standard_deviation = standard_deviation
        super(DivideStandardDeviationImage, self).__init__()

    def call(self, image):
        return image / self.standard_deviation


class ScaledResize(Processor):
    """Resizes image by returning the scales to original image.

    # Arguments
        image_size: Int, desired size of the model input.

    # Properties
        image_size: Int.

    # Methods
        call()
    """
    def __init__(self, image_size):
        self.image_size = image_size
        super(ScaledResize, self).__init__()

    def call(self, image):
        """
        # Arguments
            image: Array, raw input image.
        """
        output_image, image_scale = scale_resize(image, self.image_size)
        return output_image, image_scale
